"""Landmark Registration of functional data module.

This module contains methods to perform the landmark registration.
"""
from __future__ import annotations

import warnings
from typing import Any, Callable, Optional, Sequence, Union

import numpy as np

from ...representation import FData, FDataGrid
from ...representation.extrapolation import ExtrapolationLike
from ...representation.interpolation import SplineInterpolation
from ...typing._base import GridPointsLike
from ...typing._numpy import ArrayLike, NDArrayFloat

_FixedLocation = Union[float, Sequence[float]]
_LocationCallable = Callable[[np.ndarray], _FixedLocation]


def landmark_shift_deltas(
    fd: FData,
    landmarks: ArrayLike,
    location: Union[_FixedLocation, _LocationCallable, None] = None,
) -> NDArrayFloat:
    r"""Return the corresponding shifts to align the landmarks of the curves.

        Let :math:`t^*` the time where the landmarks of the curves will be
        aligned, and :math:`t_i` the location of the landmarks for each curve.
        The function will calculate the corresponding :math:`\delta_i` shuch
        that :math:`t_i = t^* + \delta_i`.

        This procedure will work independent of the dimension of the
        :term:`domain` and the :term:`codomain`.

    Args:
        fd: Functional data object.
        landmarks: List with the landmarks of the samples.
        location: Defines where
            the landmarks will be alligned. If a number or list is passed the
            landmarks will be alligned to it. In case of a callable is
            passed the location will be the result of the the call, the
            function should be accept as an unique parameter a numpy array
            with the list of landmarks.
            By default it will be used as location the mean of the original
            locations of the landmarks.

    Returns:
        Array containing the corresponding shifts.

    Raises:
        ValueError: If the list of landmarks does not match with the number of
            samples.

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import landmark_shift_deltas

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, random_state=1)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        >>> landmarks = landmarks.squeeze()

        The function will return the corresponding shifts

        >>> shifts = landmark_shift_deltas(fd, landmarks)
        >>> shifts.round(3)
        array([ 0.327, -0.173, -0.154])

        The registered samples can be obtained with a shift

        >>> fd.shift(shifts)
        FDataGrid(...)

    """
    landmarks = np.atleast_1d(landmarks)

    if len(landmarks) != fd.n_samples:
        raise ValueError(
            f"landmark list ({len(landmarks)}) must have the same"
            f" length than the number of samples ({fd.n_samples})",
        )

    loc_array: Union[float, Sequence[float], NDArrayFloat]

    # Parses location
    if location is None:
        loc_array = np.mean(landmarks)
    elif callable(location):
        loc_array = location(landmarks)
    else:
        loc_array = location

    loc_array = np.atleast_1d(loc_array)

    return landmarks - loc_array


def landmark_shift(
    *args: Any,
    **kwargs: Any,
) -> FDataGrid:

    warnings.warn(
        "Function 'landmark_shift' has been renamed. "
        "Use 'landmark_shift_registration' instead.",
        DeprecationWarning,
    )

    return landmark_shift_registration(*args, **kwargs)


def landmark_shift_registration(
    fd: FData,
    landmarks: ArrayLike,
    location: Union[_FixedLocation, _LocationCallable, None] = None,
    *,
    restrict_domain: bool = False,
    extrapolation: Optional[ExtrapolationLike] = None,
    grid_points: Optional[GridPointsLike] = None,
) -> FDataGrid:
    r"""
    Perform a shift of the curves to align the landmarks.

    Let :math:`t^*` the time where the landmarks of the curves will be
    aligned, :math:`t_i` the location of the landmarks for each curve
    and :math:`\delta_i= t_i - t^*`.

    The registered samples will have their feature aligned.

    .. math::
        x_i^*(t^*)=x_i(t^* + \delta_i)=x_i(t_i)

    Args:
        fd: Functional data object.
        landmarks: List with the landmarks of the samples.
        location: Defines where
            the landmarks will be alligned. If a numeric value is passed the
            landmarks will be alligned to it. In case of a callable is
            passed the location will be the result of the the call, the
            function should be accept as an unique parameter a numpy array
            with the list of landmarks.
            By default it will be used as location :math:`\frac{1}{2}(max(
            \text{landmarks})+ min(\text{landmarks}))` wich minimizes the
            max shift.
        restrict_domain: If True restricts the domain to
            avoid evaluate points outside the domain using extrapolation.
            Defaults uses extrapolation.
        extrapolation: Controls the
            extrapolation mode for elements outside the domain range.
            By default uses the method defined in fd. See extrapolation to
            more information.
        grid_points: Grid of points where
            the functions are evaluated in :func:`shift`.

    Returns:
        Functional data object with the registered samples.

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import (
        ...     landmark_shift_registration,
        ... )

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, random_state=1)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, random_state=1)
        >>> landmarks = landmarks.squeeze()

        The function will return the sample registered

        >>> landmark_shift_registration(fd, landmarks)
        FDataGrid(...)

    """
    shifts = landmark_shift_deltas(fd, landmarks, location=location)

    return fd.shift(
        shifts,
        restrict_domain=restrict_domain,
        extrapolation=extrapolation,
        grid_points=grid_points,
    )


def landmark_elastic_registration_warping(
    fd: FData,
    landmarks: ArrayLike,
    *,
    location: Optional[ArrayLike] = None,
    grid_points: Optional[GridPointsLike] = None,
) -> FDataGrid:
    """Calculate the transformation used in landmark registration.

        Let :math:`t_{ij}` the time where the sample :math:`i` has the feature
        :math:`j` and :math:`t^*_j` the new time for the feature. The warping
        function will transform the new time in the old time, i.e.,
        :math:`h_i(t^*_j)=t_{ij}`.
        The registered samples can be obtained as :math:`x^*_i(t)=x_i(h_i(t))`.

        See :footcite:`ramsay+silverman_2005_registration`
        for a detailed explanation.

    Args:
        fd: Functional data object.
        landmarks: List containing landmarks for each samples.
        location: Defines where
            the landmarks will be alligned. By default it will be used as
            location the mean of the landmarks.
        grid_points: Grid of points where
            the functions are evaluated to obtain a discrete
            representation of the object.

    Returns:
        FDataGrid with the warpings function needed to
        register the functional data object.

    Raises:
        ValueError: If the object to be registered has domain dimension greater
            than 1 or the list of landmarks or locations does not match with
            the number of samples.

    References:
        .. footbibliography::

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import (
        ...      landmark_elastic_registration_warping)

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, n_modes=2,
        ...                              random_state=9)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
        ...                                       random_state=9)
        >>> landmarks = landmarks.squeeze()

        The function will return the corresponding warping function

        >>> warping = landmark_elastic_registration_warping(fd, landmarks)
        >>> warping
        FDataGrid(...)

        The registered function can be obtained using function composition

        >>> fd.compose(warping)
        FDataGrid(...)

    """
    landmarks = np.asarray(landmarks)

    if fd.dim_domain > 1:
        raise NotImplementedError(
            "Method only implemented for objects with "
            "domain dimension up to 1.",
        )

    if len(landmarks) != fd.n_samples:
        raise ValueError(
            "The number of list of landmarks should be equal to "
            "the number of samples",
        )

    landmarks = landmarks.reshape((fd.n_samples, -1))

    location = (
        np.mean(landmarks, axis=0)
        if location is None
        else np.asarray(location)
    )

    assert isinstance(location, np.ndarray)

    n_landmarks = landmarks.shape[-1]

    data_matrix = np.empty((fd.n_samples, n_landmarks + 2))
    data_matrix[:, 0] = fd.domain_range[0][0]
    data_matrix[:, -1] = fd.domain_range[0][1]
    data_matrix[:, 1:-1] = landmarks

    if n_landmarks == len(location):
        if grid_points is None:
            grid_points = np.empty(n_landmarks + 2)
            grid_points[0] = fd.domain_range[0][0]
            grid_points[-1] = fd.domain_range[0][1]
            grid_points[1:-1] = location

    else:
        raise ValueError(
            f"Number of landmark locations should be equal than "
            f"the number of landmarks ({len(location)}) != ({n_landmarks})",
        )

    interpolation = SplineInterpolation(interpolation_order=3, monotone=True)

    warping = FDataGrid(
        data_matrix=data_matrix,
        grid_points=grid_points,
        interpolation=interpolation,
        extrapolation='bounds',
    )

    try:
        warping_points = fd.grid_points
    except AttributeError:
        warping_points = None

    return warping.to_grid(warping_points)


def landmark_registration(
    *args: Any,
    **kwargs: Any,
) -> FDataGrid:

    warnings.warn(
        "Function 'landmark_registration' has been renamed. "
        "Use 'landmark_elastic_registration' instead.",
        DeprecationWarning,
    )

    return landmark_elastic_registration(*args, **kwargs)


def landmark_elastic_registration(
    fd: FData,
    landmarks: ArrayLike,
    *,
    location: Optional[ArrayLike] = None,
    grid_points: Optional[GridPointsLike] = None,
) -> FDataGrid:
    """
    Perform landmark registration of the curves.

    Let :math:`t_{ij}` the time where the sample :math:`i` has the feature
    :math:`j` and :math:`t^*_j` the new time for the feature.
    The registered samples will have their features aligned, i.e.,
    :math:`x^*_i(t^*_j)=x_i(t_{ij})`.

    See :footcite:`ramsay+silverman_2005_registration`
    for a detailed explanation.

    Args:
        fd: Functional data object.
        landmarks: List containing landmarks for each samples.
        location: Defines where
            the landmarks will be alligned. By default it will be used as
            location the mean of the landmarks.
        grid_points: Grid of points where
            the functions are evaluated to obtain a discrete
            representation of the object. In case of objects with
            multidimensional :term:`domain` a list axis with points of
            evaluation for each dimension.

    Returns:
        FDataGrid with the functional data object registered.

    References:
        .. footbibliography::

    Examples:
        >>> from skfda.datasets import make_multimodal_landmarks
        >>> from skfda.datasets import make_multimodal_samples
        >>> from skfda.preprocessing.registration import (
        ...     landmark_elastic_registration,
        ... )
        >>> from skfda.representation.basis import BSplineBasis

        We will create a data with landmarks as example

        >>> fd = make_multimodal_samples(n_samples=3, n_modes=2,
        ...                              random_state=9)
        >>> landmarks = make_multimodal_landmarks(n_samples=3, n_modes=2,
        ...                                       random_state=9)
        >>> landmarks = landmarks.squeeze()

        The function will return the registered curves

        >>> landmark_elastic_registration(fd, landmarks)
        FDataGrid(...)

        This method will work for FDataBasis as for FDataGrids

        >>> fd = fd.to_basis(BSplineBasis(n_basis=12))
        >>> landmark_elastic_registration(fd, landmarks)
        FDataGrid(...)

    """
    warping = landmark_elastic_registration_warping(
        fd,
        landmarks,
        location=location,
        grid_points=grid_points,
    )

    return fd.to_grid(grid_points).compose(warping)



"""Preprocessing methods for functional data."""

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "dim_reduction",
        "feature_construction",
        "missing",
        "registration",
        "smoothing",
    ],
)



"""Validation utilities."""

from __future__ import annotations

import functools
import numbers
from typing import Container, Sequence, Tuple, cast

import numpy as np
from sklearn.utils import check_random_state as _check_random_state

from ..representation import FData, FDataBasis, FDataGrid
from ..typing._base import (
    DomainRange,
    DomainRangeLike,
    EvaluationPoints,
    RandomState,
    RandomStateLike,
)
from ..typing._numpy import ArrayLike


def check_fdata_dimensions(
    fd: FData,
    *,
    dim_domain: int | Container[int] | None = None,
    dim_codomain: int | Container[int] | None = None,
) -> None:
    """
    Check that a functional data object have appropriate dimensions.

    Args:
        fd: Functional data object to check.
        dim_domain: Allowed dimension(s) of the :term:`domain`.
        dim_codomain: Allowed dimension(s) of the :term:`codomain`.

    Raises:
        ValueError: If the data has not the requested dimensions.

    """
    if isinstance(dim_domain, int):
        dim_domain = {dim_domain}

    if isinstance(dim_codomain, int):
        dim_codomain = {dim_codomain}

    if dim_domain is not None:

        if fd.dim_domain not in dim_domain:
            raise ValueError(
                "Invalid domain dimension for functional data object:"
                f"{fd.dim_domain} not in {dim_domain}.",
            )

    if dim_codomain is not None:

        if fd.dim_codomain not in dim_codomain:
            raise ValueError(
                "Invalid domain dimension for functional data object:"
                f"{fd.dim_codomain} not in {dim_codomain}.",
            )


def check_fdata_same_dimensions(
    fdata1: FData,
    fdata2: FData,
) -> None:
    """
    Check that two FData have the same dimensions.

    Args:
        fdata1: First functional data object.
        fdata2: Second functional data object.

    Raises:
        ValueError: If the dimensions don't agree.

    """
    if (fdata1.dim_domain != fdata2.dim_domain):
        raise ValueError(
            f"Functional data has incompatible domain dimensions: "
            f"{fdata1.dim_domain} != {fdata2.dim_domain}",
        )

    if (fdata1.dim_codomain != fdata2.dim_codomain):
        raise ValueError(
            f"Functional data has incompatible codomain dimensions: "
            f"{fdata1.dim_codomain} != {fdata2.dim_codomain}",
        )

    if (fdata1.domain_range != fdata2.domain_range):
        raise ValueError(
            f"Functional data has incompatible domain range: "
            f"{fdata1.domain_range} != {fdata2.domain_range}",
        )


@functools.singledispatch
def _check_fdata_same_kind_specific(
    fdata1: FData,
    fdata2: FData,
) -> None:
    """Specific comparisons for subclasses."""


@_check_fdata_same_kind_specific.register
def _check_fdatagrid_same_kind_specific(
    fdata1: FDataGrid,
    fdata2: FData,
) -> None:

    # First we do an identity comparison to speed up the common case
    if fdata1.grid_points is not fdata2.grid_points:
        if not all(
            np.array_equal(g1, g2)
            for g1, g2 in zip(fdata1.grid_points, fdata2.grid_points)
        ):
            raise ValueError(
                f"Incompatible grid points between functional data objects:"
                f"{fdata1.grid_points} != {fdata2.grid_points}",
            )


@_check_fdata_same_kind_specific.register
def _check_fdatabasis_same_kind_specific(
    fdata1: FDataBasis,
    fdata2: FData,
) -> None:

    if fdata1.basis != fdata2.basis:
        raise ValueError(
            f"Incompatible basis between functional data objects:"
            f"{fdata1.basis} != {fdata2.basis}",
        )


def check_fdata_same_kind(
    fdata1: FData,
    fdata2: FData,
) -> None:
    """
    Check that two functional objects are of the same kind.

    This compares everything to ensure that the data is compatible: dimensions,
    grids, basis, when applicable. Every parameter that must be fixed for all
    samples should be the same.

    In other words: it should be possible to concatenate the two functional
    objects together.

    Args:
        fdata1: First functional data object.
        fdata2: Second functional data object.

    Raises:
        ValueError: If some attributes don't agree.

    """
    check_fdata_same_dimensions(fdata1, fdata2)

    # If the second is a subclass, reverse the order
    if isinstance(fdata2, type(fdata1)):
        fdata1, fdata2 = fdata2, fdata1

    _check_fdata_same_kind_specific(fdata1, fdata2)

    # If there is no subclassing, execute both checks
    if not isinstance(fdata1, type(fdata2)):
        _check_fdata_same_kind_specific(fdata2, fdata1)


def _valid_eval_points_shape(
    shape: Sequence[int],
    *,
    dim_domain: int,
) -> bool:
    """Check that the shape for aligned evaluation points is ok."""
    return (
        (len(shape) == 2 and shape[-1] == dim_domain)  # noqa: WPS222
        or (len(shape) <= 1 and dim_domain == 1)  # Domain ommited
        or (len(shape) == 1 and shape == (dim_domain,))  # Num. points ommited
    )


def validate_evaluation_points(
    eval_points: ArrayLike,
    *,
    aligned: bool,
    n_samples: int,
    dim_domain: int,
) -> EvaluationPoints:
    """Convert and reshape the eval_points to ndarray.

    Args:
        eval_points: Evaluation points to be reshaped.
        aligned: Boolean flag. True if all the samples
            will be evaluated at the same evaluation_points.
        n_samples: Number of observations.
        dim_domain: Dimension of the domain.

    Returns:
        Numpy array with the eval_points, if
        evaluation_aligned is True with shape `number of evaluation points`
        x `dim_domain`. If the points are not aligned the shape of the
        points will be `n_samples` x `number of evaluation points`
        x `dim_domain`.

    """
    eval_points = np.asarray(eval_points)

    shape: Sequence[int]
    if aligned:
        if _valid_eval_points_shape(eval_points.shape, dim_domain=dim_domain):
            shape = (-1, dim_domain)
        else:
            raise ValueError(
                "Invalid shape for evaluation points."
                f"An array with size (n_points, dim_domain (={dim_domain})) "
                "was expected (both can be ommited if they are 1)."
                "Instead, the received evaluation points have shape "
                f"{eval_points.shape}.",
            )
    else:
        if eval_points.shape[0] == n_samples and _valid_eval_points_shape(
            eval_points.shape[1:],
            dim_domain=dim_domain,
        ):
            shape = (n_samples, -1, dim_domain)
        else:
            raise ValueError(
                "Invalid shape for unaligned evaluation points."
                f"An array with size (n_samples (={n_samples}), "
                f"n_points, dim_domain (={dim_domain})) "
                "was expected (the last two can be ommited if they are 1)."
                "Instead, the received evaluation points have shape "
                f"{eval_points.shape}.",
            )

    return eval_points.reshape(shape)


def _validate_domain_range_limits(
    limits: Sequence[float],
) -> Tuple[float, float]:
    if len(limits) != 2 or limits[0] > limits[1]:
        raise ValueError(
            f"Invalid domain interval {limits}. "
            "Domain intervals should have 2 bounds for "
            "dimension: (lower, upper).",
        )

    lower, upper = limits
    return (float(lower), float(upper))


def validate_domain_range(domain_range: DomainRangeLike) -> DomainRange:
    """Convert sequence to a proper domain range."""
    if isinstance(domain_range[0], numbers.Real):
        domain_range = cast(Sequence[float], domain_range)
        domain_range = (domain_range,)

    domain_range = cast(Sequence[Sequence[float]], domain_range)

    return tuple(_validate_domain_range_limits(s) for s in domain_range)


def validate_random_state(random_state: RandomStateLike) -> RandomState:
    """Validate random state or seed."""
    if isinstance(random_state, np.random.Generator):
        return random_state

    return _check_random_state(random_state)  # type: ignore[no-any-return]



"""Tests for depth measures."""
import unittest

import numpy as np

import skfda
from skfda.exploratory.depth import IntegratedDepth, ModifiedBandDepth


class TestsDepthSameCurves(unittest.TestCase):
    """Test behavior when all curves in the distribution are the same."""

    def setUp(self) -> None:
        """Define the dataset."""
        data_matrix = np.tile([1, 2, 3, 4], (5, 1))

        self.fd = skfda.FDataGrid(data_matrix)

    def test_integrated_equal(self) -> None:
        """
        Test the Fraiman-Muñiz depth.

        Results should be equal to the minimum depth (0.5).

        """
        depth = IntegratedDepth()

        np.testing.assert_almost_equal(
            depth(self.fd),
            [0.5, 0.5, 0.5, 0.5, 0.5],
        )

    def test_modified_band_depth_equal(self) -> None:
        """
        Test MBD.

        Results should be equal to the maximum depth (1).

        """
        depth = ModifiedBandDepth()

        np.testing.assert_almost_equal(
            depth(self.fd),
            [1, 1, 1, 1, 1],
        )



"""Tests for FPCA."""
from __future__ import annotations

import unittest

import numpy as np

import skfda
from skfda.datasets import fetch_tecator
from skfda.misc.operators import LinearDifferentialOperator
from skfda.misc.regularization import L2Regularization
from skfda.ml.regression import FPCARegression
from skfda.representation.basis import BSplineBasis


class FPCARegressionTestCase(unittest.TestCase):
    """Tests for principal component analysis."""

    def test_fpca_reg_against_fda_usc(self) -> None:
        """Check that the results obtained are similar to those of fda.usc.

        Results obtained from fda.usc with the following R code
            library("fda.usc")
            data(tecator)
            # Fit the regression model with the first 129 observations
            x=tecator$absorp.fdata[1:129,]
            y=tecator$y$Fat[1:129]
            res2=fregre.pc(x,y,l=1:10)

            # Predict the response for the remaining observations
            n = length(tecator$y$Fat)
            xnew=tecator$absorp.fdata[130:n,]
            result = predict(res2, xnew)
            names(result) = NULL

            # Output the predicted values
            paste(
                round(result,8),
                collapse = ", "
            )
        """
        X, y = skfda.datasets.fetch_tecator(return_X_y=True, as_frame=True)
        X_train = X.iloc[:129, 0].values
        y_train = y["fat"][:129].values
        X_test = X.iloc[129:, 0].values

        fpca_regression = FPCARegression(n_components=10)

        fpca_regression.fit(X_train, y_train)
        r_predictions = np.array([  # noqa: WPS317
            24.64807256, 3.46507101, 42.8075275, 4.47183513, 5.15394474,
            8.70136459, 7.18140042, 12.26860992, 11.89761155, 11.43831823,
            21.70303956, 25.75704241, 30.7086502, 48.10692261, 10.6258519,
            4.07226986, 5.3564939, 7.34333975, 6.02196557, 7.17052614,
            8.96339574, 7.48137304, 8.80638685, 9.91950449, 9.61385587,
            10.2969699, 9.72478128, 10.02837667, 10.72087154, 15.50611227,
            13.75408906, 15.45072713, 15.77815535, 16.4328961, 21.82526526,
            25.98049031, 26.75486697, 25.45727954, 31.41793557, 28.88122152,
            29.93498123, 44.2845357, 54.8070113, 43.58988546, 18.35395886,
            6.77862943, 4.07226986, 9.00970654, 9.02405569, 8.25587234,
            12.1732419, 14.28128606, 16.82891438, 24.1552859, 29.6879001,
            34.1931269, 54.85800891, 11.0114267, 8.5458698, 7.80994054,
            6.58579322, 7.24964726, 8.85435095, 7.97581994, 7.85785756,
            7.83365218, 8.4137662, 9.33079262, 9.78579099, 10.19765504,
            11.24655925, 12.18472769, 13.15521698, 15.01021835,
            11.43831823, 16.05351555, 17.47985504, 19.86368787,
            23.8272112, 26.88801046, 26.79250389, 28.7660723,
            31.95784747, 34.81938152, 42.52653123, 54.85800891,
        ])

        predictions = fpca_regression.predict(X_test)

        np.testing.assert_allclose(r_predictions, predictions, rtol=5e-3)

    def test_fpca_reg_against_fda_usc_reg(self) -> None:
        """Check that the results obtained are similar to those of fda.usc.

        Results obtained from fda.usc with the following R code
            library("fda.usc")
            data(tecator)
            # Fit the regression model with the first 129 observations
            x=tecator$absorp.fdata[1:129,]
            y=tecator$y$Fat[1:129]
            res2=fregre.pc(x,y,l=1:10, lambda = 1, P=c(0,0,1))

            # Predict the response for the remaining observations
            n = length(tecator$y$Fat)
            xnew=tecator$absorp.fdata[130:n,]
            result = predict(res2, xnew)
            names(result) = NULL

            # Output the predicted values
            paste(
                round(result,8),
                collapse = ", "
            )
        """
        X, y = skfda.datasets.fetch_tecator(return_X_y=True, as_frame=True)
        X_train = X.iloc[:129, 0].values
        y_train = y["fat"][:129].values
        X_test = X.iloc[129:, 0].values
        pen_order = 2

        # Two correction factors are needed to obtain the same results as
        # fda.usc

        # The first factor compensates for the fact that the difference
        # matrices in fda.usc are scaled by the mean of the deltas
        # between grid points. This diference is introduced in
        # the function D.penalty (fdata2pc.R:479) in fda.usc.
        grid_points = X_test[0].grid_points[0]
        grid_step = np.mean(np.diff(grid_points))
        factor1 = grid_step**(2 * pen_order - 1)

        # The second factor compensates a scaling done to the final
        # regularization matrix in fda.usc. This scaling is introduced
        # in the function fregre.pc (fregre.pc.R:165) in fda.usc.
        factor2 = np.diff(X_train.domain_range[0]) / len(grid_points)
        factor2 **= (2 * pen_order + 1)

        # The total factor that we need to apply to the regularization
        # parameter of the regression is the product of the two factors
        total_factor = factor1 * factor2

        fpca_regression = FPCARegression(
            n_components=10,
            pca_regularization=L2Regularization(
                LinearDifferentialOperator(pen_order),
                regularization_parameter=1.0,
            ),
            regression_regularization=L2Regularization(
                LinearDifferentialOperator(pen_order),
                regularization_parameter=1.0 * total_factor,  # noqa: WPS345
            ),
        )

        fpca_regression.fit(X_train, y_train)
        r_predictions = np.array([  # noqa: WPS317
            23.48391439, 8.45397463, 43.44474891, 6.0981306, 6.56256704,
            8.7670954, 6.22132003, 12.8571075, 14.11658575, 13.57439588,
            23.96133718, 26.81618552, 30.31754827, 47.34895476, 11.53618091,
            5.49448547, 5.60875954, 8.57410351, 6.41573393, 7.22929637,
            7.53334465, 6.68773565, 8.6818259, 10.74793989, 8.6782855,
            11.89552059, 10.40896226, 10.93569607, 11.87369452, 14.57111824,
            14.5522881, 13.64065941, 16.03304023, 15.60254571, 21.11275437,
            25.03334373, 26.57624209, 25.44579176, 31.43561996, 28.03967631,
            29.77583234, 41.44325753, 54.14277276, 44.0569423, 22.90209546,
            8.85428988, 5.49448547, 9.36725951, 7.86510101, 6.22786929,
            13.71579956, 15.88516382, 18.05432943, 23.13267508, 28.48278103,
            36.59999903, 54.07821032, 12.55490093, 9.85357179, 8.79073997,
            6.95827529, 8.00027917, 9.19686289, 9.35734081, 8.60987473,
            7.81116661, 8.57285562, 7.73267045, 10.19060339, 9.96290068,
            12.57528552, 11.601402, 12.22342486, 14.19036667, 13.57439588,
            16.14576329, 15.87859389, 19.76979144, 23.62449942, 26.47268082,
            26.14126515, 27.52545896, 30.58823007, 33.62623406, 40.96307478,
            54.07821032,
        ])

        predictions = fpca_regression.predict(X_test)

        # The tolerance is not lower because the results are not
        # exactly the same as in fda.usc. The components calculated
        # by fda.usc are not exactly the same (they are orthogonal).
        # Additionaly, the penalization matrices are calculated using
        # a different method.
        np.testing.assert_allclose(r_predictions, predictions, rtol=1.5e-2)

    def test_fpca_reg_basis_vs_grid(self):
        """
        Compare results between grid and basis.

        Compare that the end result is the same when using the grid or
        basis representation of the data.
        """
        X, y = fetch_tecator(return_X_y=True)
        X = X.to_basis(BSplineBasis(n_basis=20))
        X_train = X[:129]
        X_test = X[129:]

        sampling_grid = np.linspace(
            X.domain_range[0][0], X.domain_range[0][1], 100,
        )

        X_grid = X.to_grid(grid_points=sampling_grid)
        X_grid_train = X_grid[:129]
        X_grid_test = X_grid[129:]

        fpca_regression = FPCARegression(
            n_components=10,
            pca_regularization=L2Regularization(
                LinearDifferentialOperator(2), regularization_parameter=10,
            ),
            regression_regularization=L2Regularization(
                LinearDifferentialOperator(2), regularization_parameter=10,
            ),
        )

        fpca_regression.fit(X_train, y[:129, 0])
        predictions_basis = fpca_regression.predict(X_test)

        fpca_regression.fit(X_grid_train, y[:129, 0])
        predictions_grid = fpca_regression.predict(X_grid_test)

        np.testing.assert_allclose(
            predictions_basis,
            predictions_grid,
            rtol=6e-3,
        )


if __name__ == "__main__":
    unittest.main()



"""Tests for FDataBasis evaluation."""
import unittest

import numpy as np

from skfda.representation.basis import (
    BSplineBasis,
    ConstantBasis,
    FDataBasis,
    FourierBasis,
    MonomialBasis,
    TensorBasis,
    VectorValuedBasis,
)


class TestFDataBasisEvaluation(unittest.TestCase):
    """
    Test FDataBasis evaluation.

    This class includes tests that don't depend on the particular based used.

    """

    def test_evaluation_single_point(self) -> None:
        """Test the evaluation of a single point FDataBasis."""
        fourier = FourierBasis(
            domain_range=(0, 1),
            n_basis=3,
        )

        coefficients = np.array([
            [0.00078238, 0.48857741, 0.63971985],
            [0.01778079, 0.73440271, 0.20148638],
        ])

        f = FDataBasis(fourier, coefficients)

        # Test different ways of call f with a point
        res = np.array(
            [-0.903918107989282, -0.267163981229459],
        ).reshape((2, 1, 1))

        np.testing.assert_allclose(f([0.5]), res)
        np.testing.assert_allclose(f((0.5,)), res)
        np.testing.assert_allclose(f(0.5), res)
        np.testing.assert_allclose(f(np.array([0.5])), res)
        np.testing.assert_allclose(f(np.array(0.5)), res)

    def test_evaluation_grid_1d(self) -> None:
        """
        Test the evaluation of FDataBasis with grid=True.

        Nothing should change as the domain dimension is 1.

        """
        fourier = FourierBasis(
            domain_range=(0, 1),
            n_basis=3,
        )

        coefficients = np.array([
            [0.00078238, 0.48857741, 0.63971985],
            [0.01778079, 0.73440271, 0.20148638],
        ])

        f = FDataBasis(fourier, coefficients)
        t = np.linspace(0, 1, 4)

        res_test = f(t)

        # Different ways to pass the axes
        np.testing.assert_allclose(
            f(t, grid=True),
            res_test,
        )
        np.testing.assert_allclose(
            f((t,), grid=True),
            res_test,
        )
        np.testing.assert_allclose(
            f([t], grid=True),
            res_test,
        )
        np.testing.assert_allclose(
            f(np.atleast_2d(t), grid=True),
            res_test,
        )

        # Number of axis different than the domain dimension (1)
        with np.testing.assert_raises(ValueError):
            f((t, t), grid=True)

    def test_evaluation_unaligned(self) -> None:
        """Test the unaligned evaluation."""
        fourier = FourierBasis(
            domain_range=(0, 1),
            n_basis=3,
        )

        coefficients = np.array([
            [0.00078238, 0.48857741, 0.63971985],
            [0.01778079, 0.73440271, 0.20148638],
        ])

        f = FDataBasis(fourier, coefficients)
        t = np.linspace(0, 1, 4)

        # Test same result than normal evaluation
        np.testing.assert_allclose(
            f([1]),
            f([[1], [1]], aligned=False),
        )
        np.testing.assert_allclose(
            f(t),
            f(np.vstack((t, t)), aligned=False),
        )

        # Different evaluation times
        t_multiple = [[0, 0.5], [0.2, 0.7]]
        np.testing.assert_allclose(
            f(t_multiple[0])[0],
            f(t_multiple, aligned=False)[0],
        )

        np.testing.assert_allclose(
            f(t_multiple[1])[1],
            f(t_multiple, aligned=False)[1],
        )


class TestBasisEvaluationFourier(unittest.TestCase):
    """Test FDataBasis with Fourier basis."""

    def test_simple_evaluation(self) -> None:
        """Compare Fourier evaluation with R package fda."""
        fourier = FourierBasis(
            domain_range=(0, 2),
            n_basis=5,
        )

        coefficients = np.array([
            [1, 2, 3, 4, 5],
            [6, 7, 8, 9, 10],
        ])

        f = FDataBasis(fourier, coefficients)

        t = np.linspace(0, 2, 11)

        # Results in R package fda
        res = np.array([
            [  # noqa: WPS317
                8.71, 9.66, 1.84, -4.71, -2.80, 2.71,
                2.45, -3.82, -6.66, -0.30, 8.71,
            ],
            [  # noqa: WPS317
                22.24, 26.48, 10.57, -4.95, -3.58, 6.24,
                5.31, -7.69, -13.32, 1.13, 22.24,
            ],
        ])[..., np.newaxis]

        np.testing.assert_allclose(f(t), res, atol=1e-2)

    def test_evaluation_derivative(self) -> None:
        """Test the evaluation of the derivative of Fourier."""
        fourier = FourierBasis(domain_range=(0, 1), n_basis=3)

        coefficients = np.array([
            [0.00078238, 0.48857741, 0.63971985],
            [0.01778079, 0.73440271, 0.20148638],
        ])

        f = FDataBasis(fourier, coefficients)

        t = np.linspace(0, 1, 4)

        res = np.array([  # noqa: WPS317
            4.34138447771721, -7.09352774867064, 2.75214327095343,
            4.34138447771721, 6.52573053999253,
            -4.81336320468984, -1.7123673353027, 6.52573053999253,
        ]).reshape((2, 4, 1))

        f_deriv = f.derivative()
        np.testing.assert_allclose(
            f_deriv(t),
            res,
        )


class TestBasisEvaluationBSpline(unittest.TestCase):
    """Test FDataBasis with B-spline basis."""

    def test_simple_evaluation(self) -> None:
        """Compare BSpline evaluation with R package fda."""
        bspline = BSplineBasis(
            domain_range=(0, 2),
            n_basis=5,
        )

        coefficients = np.array([
            [1, 2, 3, 4, 5],
            [6, 7, 8, 9, 10],
        ])

        f = FDataBasis(bspline, coefficients)

        t = np.linspace(0, 2, 11)

        # Results in R package fda
        res = np.array([
            [  # noqa: WPS317
                1, 1.54, 1.99, 2.37, 2.7, 3,
                3.3, 3.63, 4.01, 4.46, 5,
            ],
            [  # noqa: WPS317
                6, 6.54, 6.99, 7.37, 7.7, 8,
                8.3, 8.63, 9.01, 9.46, 10,
            ],
        ])[..., np.newaxis]

        np.testing.assert_allclose(f(t), res, atol=1e-2)

    def test_evaluation_derivative(self) -> None:
        """Test the evaluation of the derivative of BSpline."""
        bspline = BSplineBasis(
            domain_range=(0, 1),
            n_basis=5,
            order=3,
        )

        coefficients = [
            [0.00078238, 0.48857741, 0.63971985, 0.23, 0.33],
            [0.01778079, 0.73440271, 0.20148638, 0.54, 0.12],
        ]

        f = FDataBasis(bspline, coefficients)

        t = np.linspace(0, 1, 4)

        f_deriv = f.derivative()
        np.testing.assert_allclose(
            f_deriv(t).round(3),
            np.array([
                [2.927, 0.453, -1.229, 0.6],
                [4.3, -1.599, 1.016, -2.52],
            ])[..., np.newaxis],
        )


class TestBasisEvaluationMonomial(unittest.TestCase):
    """Test FDataBasis with Monomial basis."""

    def test_evaluation_simple_monomial(self) -> None:
        """Compare Monomial evaluation with R package fda."""
        monomial = MonomialBasis(
            domain_range=(0, 2),
            n_basis=5,
        )

        coefficients = np.array([
            [1, 2, 3, 4, 5],
            [6, 7, 8, 9, 10],
        ])

        f = FDataBasis(monomial, coefficients)

        t = np.linspace(0, 2, 11)

        # Results in R package fda
        res = np.array([
            [  # noqa: WPS317
                1.00, 1.56, 2.66, 4.79, 8.62, 15.00,
                25.00, 39.86, 61.03, 90.14, 129.00,
            ],
            [  # noqa: WPS317
                6.00, 7.81, 10.91, 16.32, 25.42, 40.00,
                62.21, 94.59, 140.08, 201.98, 284.00,
            ],
        ])[..., np.newaxis]

        np.testing.assert_allclose(f(t), res, atol=1e-2)

    def test_evaluation_derivative(self) -> None:
        """Test the evaluation of the derivative of Monomial."""
        monomial = MonomialBasis(
            domain_range=(0, 1),
            n_basis=3,
        )

        coefficients = [
            [1.0, 2.0, 3.0],
            [0.5, 1.4, 1.3],
        ]

        f = FDataBasis(monomial, coefficients)

        t = np.linspace(0, 1, 4)

        f_deriv = f.derivative()
        np.testing.assert_allclose(
            f_deriv(t).round(3),
            np.array([
                [2.0, 4.0, 6.0, 8.0],
                [1.4, 2.267, 3.133, 4.0],
            ])[..., np.newaxis],
        )


class TestBasisEvaluationVectorValued(unittest.TestCase):
    """Test basis for vector-valued functions."""

    def test_constant(self) -> None:
        """Test vector-valued constant basis."""
        basis_first = ConstantBasis()
        basis_second = ConstantBasis()

        basis = VectorValuedBasis([basis_first, basis_second])

        fd = FDataBasis(basis=basis, coefficients=[[1, 2], [3, 4]])

        self.assertEqual(fd.dim_codomain, 2)

        res = np.array([[[1, 2]], [[3, 4]]])

        np.testing.assert_allclose(fd(0), res)

    def test_monomial(self) -> None:
        """Test vector-valued monomial basis."""
        basis_first = ConstantBasis(domain_range=(0, 5))
        basis_second = MonomialBasis(
            n_basis=3,
            domain_range=(0, 5),
        )

        basis = VectorValuedBasis([basis_first, basis_second])

        fd = FDataBasis(
            basis=basis,
            coefficients=[
                [1, 2, 3, 4],
                [3, 4, 5, 6],
            ],
        )

        self.assertEqual(fd.dim_codomain, 2)

        np.testing.assert_allclose(fd.domain_range[0], (0, 5))

        res = np.array([
            [[1, 2], [1, 9], [1, 24]],
            [[3, 4], [3, 15], [3, 38]],
        ])

        np.testing.assert_allclose(fd([0, 1, 2]), res)


class TestBasisEvaluationTensor(unittest.TestCase):
    """Test tensor basis for multivariable functions."""

    def test_tensor_monomial_constant(self) -> None:
        """Test monomial ⊗ constant basis."""
        basis = TensorBasis([
            MonomialBasis(n_basis=2),
            ConstantBasis(),
        ])

        fd = FDataBasis(
            basis=basis,
            coefficients=[1, 1],
        )

        self.assertEqual(fd.dim_domain, 2)
        self.assertEqual(fd.dim_codomain, 1)

        np.testing.assert_allclose(fd([0, 0]), [[[1]]])

        np.testing.assert_allclose(fd([0.5, 0.5]), [[[1.5]]])

        np.testing.assert_allclose(
            fd([(0, 0), (0.5, 0.5)]),
            [[[1.0], [1.5]]],
        )

        fd_grid = fd.to_grid()

        fd2 = fd_grid.to_basis(basis)

        np.testing.assert_allclose(fd.coefficients, fd2.coefficients)


if __name__ == '__main__':
    unittest.main()



"""Common types."""
from typing import Optional, Sequence, Tuple, TypeVar, Union

import numpy as np
from typing_extensions import Protocol

from ._numpy import ArrayLike, NDArrayFloat

VectorType = TypeVar("VectorType")

DomainRange = Tuple[Tuple[float, float], ...]
DomainRangeLike = Union[
    DomainRange,
    Sequence[float],
    Sequence[Sequence[float]],
]

LabelTuple = Tuple[Optional[str], ...]
LabelTupleLike = Sequence[Optional[str]]

GridPoints = Tuple[NDArrayFloat, ...]
GridPointsLike = Union[ArrayLike, Sequence[ArrayLike]]

EvaluationPoints = NDArrayFloat


RandomStateLike = Union[int, np.random.RandomState, np.random.Generator, None]
RandomState = Union[np.random.RandomState, np.random.Generator]


class Vector(Protocol):
    """
    Protocol representing a generic vector.

    It should accept numpy arrays and FData, among other things.
    """

    def __add__(
        self: VectorType,
        __other: VectorType,  # noqa: WPS112
    ) -> VectorType:
        pass

    def __sub__(
        self: VectorType,
        __other: VectorType,  # noqa: WPS112
    ) -> VectorType:
        pass

    def __mul__(
        self: VectorType,
        __other: float,  # noqa: WPS112
    ) -> VectorType:
        pass



"""
Module to interpolate functional data objects.
"""
from __future__ import annotations

import abc
from typing import TYPE_CHECKING, Any, Callable, Sequence

import numpy as np
from scipy.interpolate import (
    PchipInterpolator,
    RegularGridInterpolator,
    make_interp_spline,
)

from ..typing._base import EvaluationPoints
from ..typing._numpy import ArrayLike, NDArrayFloat
from .evaluator import Evaluator

if TYPE_CHECKING:
    from .grid import FDataGrid


class _BaseInterpolation(Evaluator):
    """ABC for interpolations."""

    @abc.abstractmethod
    def _evaluate_aligned(
        self,
        fdata: FDataGrid,
        eval_points: EvaluationPoints,
    ) -> NDArrayFloat:
        """Evaluate at aligned points."""
        pass

    def _evaluate_unaligned(
        self,
        fdata: FDataGrid,
        eval_points: EvaluationPoints,
    ) -> NDArrayFloat:
        """Evaluate at unaligned points."""
        return np.vstack([
            self._evaluate_aligned(f, e)
            for f, e in zip(fdata, eval_points)
        ])

    def _evaluate(  # noqa: D102
        self,
        fdata: FDataGrid,
        eval_points: ArrayLike,
        *,
        aligned: bool = True,
    ) -> NDArrayFloat:

        from ..misc.validation import validate_evaluation_points

        eval_points = validate_evaluation_points(
            eval_points,
            aligned=aligned,
            n_samples=fdata.n_samples,
            dim_domain=fdata.dim_domain,
        )

        return (
            self._evaluate_aligned(fdata, eval_points)
            if aligned
            else self._evaluate_unaligned(fdata, eval_points)
        )


class _RegularGridInterpolatorWrapper():

    def __init__(
        self,
        fdatagrid: FDataGrid,
        interpolation_order: int,
    ) -> None:
        self.fdatagrid = fdatagrid
        self.interpolation_order = interpolation_order

        if self.interpolation_order == 0:
            method = 'nearest'
        elif self.interpolation_order == 1:
            method = 'linear'
        elif self.interpolation_order == 3:
            method = 'cubic'
        elif self.interpolation_order == 5:
            method = 'quintic'
        else:
            raise ValueError(
                f"Invalid interpolation order: {self.interpolation_order}.",
            )
        self.interpolator = RegularGridInterpolator(
            self.fdatagrid.grid_points,
            np.moveaxis(self.fdatagrid.data_matrix, 0, -2),
            method=method,
            bounds_error=False,
            fill_value=None,
        )

    def __call__(
        self,
        eval_points: EvaluationPoints,
    ) -> NDArrayFloat:
        return np.moveaxis(
            self.interpolator(eval_points),
            0,
            1,
        )


class SplineInterpolation(_BaseInterpolation):
    """
    Spline interpolation.

    Spline interpolation of discretized functional objects. Implements
    different interpolation methods based in splines, using the sample
    points of the grid as nodes to interpolate.

    See the interpolation example to a detailled explanation.

    Attributes:
        interpolation_order (int, optional): Order of the interpolation, 1
            for linear interpolation, 2 for cuadratic, 3 for cubic and so
            on. In case of curves and surfaces there is available
            interpolation up to degree 5. For higher dimensional objects
            only linear or nearest interpolation is available. Default
            lineal interpolation.
        smoothness_parameter (float, optional): Penalisation to perform
            smoothness interpolation. Option only available for curves and
            surfaces. If 0 the residuals of the interpolation will be 0.
            Defaults 0.
        monotone (boolean, optional): Performs monotone interpolation in
            curves using a PCHIP interpolator. Only valid for curves (domain
            dimension equal to 1) and interpolation order equal to 1 or 3.
            Defaults false.

    """

    def __init__(
        self,
        interpolation_order: int = 1,
        *,
        monotone: bool = False,
    ) -> None:
        self.interpolation_order = interpolation_order
        self.monotone = monotone

    def _get_interpolator_1d(
        self,
        fdatagrid: FDataGrid,
    ) -> Callable[[EvaluationPoints], NDArrayFloat]:
        if (
            isinstance(self.interpolation_order, Sequence)
            or not 1 <= self.interpolation_order <= 5
        ):
            raise ValueError(
                f"Invalid degree of interpolation "
                f"({self.interpolation_order}). Must be "
                f"an integer greater than 0 and lower or "
                f"equal than 5.",
            )

        if self.monotone and self.interpolation_order not in {1, 3}:
            raise ValueError(
                f"monotone interpolation of degree "
                f"{self.interpolation_order}"
                f"not supported.",
            )

        # Monotone interpolation of degree 1 is performed with linear spline
        monotone = self.monotone and self.interpolation_order > 1

        if monotone:
            return PchipInterpolator(  # type: ignore[no-any-return]
                fdatagrid.grid_points[0],
                fdatagrid.data_matrix,
                axis=1,
            )

        return make_interp_spline(  # type: ignore[no-any-return]
            fdatagrid.grid_points[0],
            fdatagrid.data_matrix,
            k=self.interpolation_order,
            axis=1,
            # Orders 0 and 1 behave well
            check_finite=self.interpolation_order > 1,
        )

    def _get_interpolator_nd(
        self,
        fdatagrid: FDataGrid,
    ) -> Callable[[EvaluationPoints], NDArrayFloat]:

        return _RegularGridInterpolatorWrapper(
            fdatagrid,
            interpolation_order=self.interpolation_order,
        )

    def _get_interpolator(
        self,
        fdatagrid: FDataGrid,
    ) -> Callable[[EvaluationPoints], NDArrayFloat]:

        if fdatagrid.dim_domain == 1:
            return self._get_interpolator_1d(fdatagrid)

        elif self.monotone:
            raise ValueError(
                "Monotone interpolation is only supported with "
                "domain dimension equal to 1.",
            )

        return self._get_interpolator_nd(fdatagrid)

    def _evaluate_aligned(  # noqa: D102
        self,
        fdata: FDataGrid,
        eval_points: EvaluationPoints,
    ) -> NDArrayFloat:

        interpolator = self._get_interpolator(fdata)
        return np.reshape(
            interpolator(eval_points),
            (fdata.n_samples, -1, fdata.dim_codomain),
        )

    def __repr__(self) -> str:
        return (
            f"{type(self).__name__}("
            f"interpolation_order={self.interpolation_order}, "
            f"monotone={self.monotone})"
        )

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and self.interpolation_order == other.interpolation_order
            and self.monotone == other.monotone
        )



"""Functional Transformers Module."""

from __future__ import annotations

import itertools
from typing import Optional, Sequence, Tuple, TypeVar, Union, cast

import numpy as np
from typing_extensions import Literal, TypeGuard

from ..._utils.ndfunction._functions import (
    _average_function_ufunc,
    _UnaryUfunc,
)
from ...misc.validation import check_fdata_dimensions
from ...representation import FData, FDataBasis, FDataGrid
from ...typing._base import DomainRangeLike
from ...typing._numpy import ArrayLike, NDArrayBool, NDArrayFloat, NDArrayInt

T = TypeVar("T", bound=Union[NDArrayFloat, FDataGrid])


def _sequence_of_ints(data: Sequence[object]) -> TypeGuard[Sequence[int]]:
    """Check that every element is an integer."""
    return all(isinstance(d, int) for d in data)


def local_averages(
    data: FData,
    *,
    domains: int | Sequence[int] | Sequence[DomainRangeLike],
) -> NDArrayFloat:
    r"""
    Calculate the local averages of given data in the desired domains.

    It takes functional data and performs the following map:

    .. math::
        f_1(X) = \frac{1}{|T_1|} \int_{T_1} X(t) dt,\dots, \\
        f_p(X) = \frac{1}{|T_p|} \int_{T_p} X(t) dt

    where :math:`T_1, \dots, T_p` are subregions of the original
    :term:`domain`.

    Args:
        data: FDataGrid or FDataBasis where we want to
            calculate the local averages.
        domains: Domains for each local average. It is possible to
            pass a number or a list of numbers to automatically split
            each dimension in that number of intervals and use them for
            the averages.

    Returns:
        ndarray of shape (n_samples, n_domains, codomain_dimension) with
        the transformed data.

    See also:
        :class:`~skfda.preprocessing.feature_construction.LocalAveragesTransformer`

    Examples:
        We import the Berkeley Growth Study dataset.
        We will use only the first 3 samples to make the
        example easy

        >>> from skfda.datasets import fetch_growth
        >>> dataset = fetch_growth(return_X_y=True)[0]
        >>> X = dataset[:3]

        We can choose the intervals used for the local averages. For example,
        we could in this case use the averages at different stages of
        development of the child: from 1 to 3 years, from 3 to 10 and from
        10 to 18:

        >>> import numpy as np
        >>> from skfda.preprocessing.feature_construction import local_averages
        >>> averages = local_averages(
        ...     X,
        ...     domains=[(1, 3), (3, 10), (10, 18)],
        ... )
        >>> np.around(averages, decimals=2)
        array([[[  91.37],
                [ 126.52],
                [ 179.02]],
               [[  87.51],
                [ 120.71],
                [ 158.81]],
               [[  86.36],
                [ 115.04],
                [ 156.37]]])

        A different possibility is to decide how many intervals we want to
        consider.  For example, we could want to split the domain in 2
        intervals of the same length.

        >>> np.around(local_averages(X, domains=2), decimals=2)
        array([[[ 116.94],
                [ 177.26]],
               [[ 111.86],
                [ 157.62]],
               [[ 107.29],
                [ 154.97]]])
    """
    if isinstance(domains, int):
        domains = [domains] * data.dim_domain

    if _sequence_of_ints(domains):
        # Get a list of arrays with the interval endpoints
        interval_endpoints = [
            np.linspace(
                domain_range[0],
                domain_range[1],
                num=n_intervals + 1,
            )
            for n_intervals, domain_range in zip(domains, data.domain_range)
        ]

        # Get all combinations of intervals as ranges
        domains = list(
            itertools.product(
                *[zip(p, p[1:]) for p in interval_endpoints],
            ),
        )

    domains = cast(Sequence[DomainRangeLike], domains)

    integrated_data = [
        unconditional_expected_value(
            data,
            lambda x: x,
            domain=domain,
        )
        for domain in domains
    ]
    return np.swapaxes(integrated_data, 0, 1)


def _calculate_curves_occupation(
    curve_y_coordinates: NDArrayFloat,
    curve_x_coordinates: NDArrayFloat,
    intervals: Sequence[Tuple[float, float]],
) -> NDArrayFloat:

    y1, y2 = np.asarray(intervals).T

    if any(np.greater(y1, y2)):
        raise ValueError(
            "Interval limits (a,b) should satisfy a <= b.",
        )

    # Reshape original curves so they have one dimension less
    curve_y_coordinates = curve_y_coordinates[:, :, 0]

    # Calculate interval sizes on the X axis
    intervals_x_axis = curve_x_coordinates[1:] - curve_x_coordinates[:-1]

    # Calculate which points are inside the interval given (y1,y2) on Y axis
    greater_than_y1 = (
        curve_y_coordinates[:, :, np.newaxis]
        >= y1[np.newaxis, np.newaxis, :]
    )
    less_than_y2 = (
        curve_y_coordinates[:, :, np.newaxis]
        <= y2[np.newaxis, np.newaxis, :]
    )
    inside_interval_bools = greater_than_y1 & less_than_y2

    # Calculate intervals on X axis where the points are inside Y axis interval
    intervals_x_inside = (
        inside_interval_bools
        * intervals_x_axis[:, np.newaxis]
    )

    return np.sum(intervals_x_inside, axis=1)  # type: ignore[no-any-return]


def occupation_measure(
    data: FData,
    intervals: Sequence[Tuple[float, float]],
    *,
    n_points: Optional[int] = None,
) -> NDArrayFloat:
    r"""
    Calculate the occupation measure of a grid.

    It performs the following map:

    ..math:
        :math:`f_1(X) = |t: X(t)\in T_p|,\dots,|t: X(t)\in T_p|`

    where :math:`{T_1,\dots,T_p}` are disjoint intervals in
    :math:`\mathbb{R}` and | | stands for the Lebesgue measure.

    The calculations are based on evaluation at a grid of points. In case of
    :class:`FDataGrid` the original grid is taken unless ``n_points`` is
    specified. In case of :class:`FDataBasis` it is mandatory to pass the
    number of points. If the result of this function is not accurate enough
    try to increase the grid of points.

    Args:
        data: Functional data where we want to calculate the occupation
            measure.
        intervals: ndarray of tuples containing the
            intervals we want to consider. The shape should be
            (n_sequences,2)
        n_points: Number of points to evaluate in the domain.
            By default will be used the points defined on the FDataGrid.
            On a FDataBasis this value should be specified.

    Returns:
        ndarray of shape (n_samples, n_intervals)
        with the transformed data.

    Examples:
        We will create the FDataGrid that we will use to extract
        the occupation measure.

        >>> from skfda.representation import FDataGrid
        >>> import numpy as np
        >>> t = np.linspace(0, 10, 100)
        >>> fd_grid = FDataGrid(
        ...     data_matrix=[
        ...         t,
        ...         2 * t,
        ...         np.sin(t),
        ...     ],
        ...     grid_points=t,
        ... )

        Finally we call to the occupation measure function with the
        intervals that we want to consider. In our case (0.0, 1.0)
        and (2.0, 3.0). We need also to specify the number of points
        we want that the function takes into account to interpolate.
        We are going to use 501 points.

        >>> from skfda.preprocessing.feature_construction import (
        ...     occupation_measure,
        ... )
        >>> np.around(
        ...     occupation_measure(
        ...         fd_grid,
        ...         [(0.0, 1.0), (2.0, 3.0)],
        ...         n_points=501,
        ...     ),
        ...     decimals=2,
        ... )
        array([[ 0.98,  1.  ],
               [ 0.5 ,  0.52],
               [ 6.28,  0.  ]])

    """
    if isinstance(data, FDataBasis) and n_points is None:
        raise ValueError(
            "Number of points to consider, should be given "
            + " as an argument for a FDataBasis. Instead None was passed.",
        )

    check_fdata_dimensions(
        data,
        dim_domain=1,
        dim_codomain=1,
    )

    if n_points is None:
        function_x_coordinates = data.grid_points[0]
        function_y_coordinates = data.data_matrix
    else:
        function_x_coordinates = np.linspace(
            data.domain_range[0][0],
            data.domain_range[0][1],
            num=n_points,
        )
        function_y_coordinates = data(function_x_coordinates[1:])

    return _calculate_curves_occupation(  # noqa: WPS317
        function_y_coordinates,
        function_x_coordinates,
        intervals,
    )


def number_crossings(
    fd: FDataGrid,
    *,
    levels: ArrayLike = 0,
    direction: Literal["up", "down", "all"] = "all",
) -> NDArrayInt:
    r"""
    Calculate the number of crossings to a level of a FDataGrid.

    Let f_1(X) = N_i, where N_i is the number of up crossings of X
    to a level c_i \in \mathbb{R}, i = 1,\dots,p.

    Recall that the process X(t) is said to have an up crossing of c
    at :math:`t_0 > 0` if for some :math:`\epsilon >0`, X(t) $\leq$
    c if t :math:'\in (t_0 - \epsilon, t_0) and X(t) $\geq$ c if
    :math:`t\in (t_0, t_0+\epsilon)`.

    If the trajectories are differentiable, then
    :math:`N_i = card\{t \in[a,b]: X(t) = c_i, X' (t) > 0\}.`

        Args:
            fd: FDataGrid where we want to calculate
                the number of up crossings.
            levels: Sequence of numbers including the levels
                we want to consider for the crossings. By
                default it calculates zero-crossings.
            direction: Whether to consider only up-crossings,
                down-crossings or both.

        Returns:
            ndarray of shape (n_samples, len(levels))\
            with the values of the counters.

    Examples:
        For this example we will use a well known function so the correct
        functioning of this method can be checked.
        We will create and use a DataFrame with a sample extracted from
        the Bessel Function of first type and order 0.
        First of all we import the Bessel Function and create the X axis
        data grid. Then we create the FdataGrid.

        >>> from skfda.preprocessing.feature_construction import (
        ...     number_crossings,
        ... )
        >>> from scipy.special import jv
        >>> import numpy as np
        >>> x_grid = np.linspace(0, 14, 14)
        >>> fd_grid = FDataGrid(
        ...     data_matrix=[jv([0], x_grid)],
        ...     grid_points=x_grid,
        ... )
        >>> fd_grid.data_matrix
        array([[[ 1.        ],
                [ 0.73041066],
                [ 0.13616752],
                [-0.32803875],
                [-0.35967936],
                [-0.04652559],
                [ 0.25396879],
                [ 0.26095573],
                [ 0.01042895],
                [-0.22089135],
                [-0.2074856 ],
                [ 0.0126612 ],
                [ 0.20089319],
                [ 0.17107348]]])

        Finally we evaluate the number of up crossings method with the
        FDataGrid created.

        >>> number_crossings(fd_grid, levels=0, direction="up")
        array([[2]])
    """
    # This is only defined for univariate functions
    check_fdata_dimensions(fd, dim_domain=1, dim_codomain=1)

    levels = np.atleast_1d(levels)
    curves = fd.data_matrix[:, :, 0]

    distances = np.subtract.outer(levels, curves)

    points_greater = distances >= 0
    points_smaller = distances <= 0

    growing = distances[:, :, :-1] < points_greater[:, :, 1:]
    lowering = distances[:, :, :-1] > points_greater[:, :, 1:]

    upcrossing_positions: NDArrayBool = (
        points_smaller[:, :, :-1] & points_greater[:, :, 1:] & growing
    )

    downcrossing_positions: NDArrayBool = (
        points_greater[:, :, :-1] & points_smaller[:, :, 1:] & lowering
    )

    positions = {
        "all": upcrossing_positions | downcrossing_positions,
        "up": upcrossing_positions,
        "down": downcrossing_positions,
    }

    return np.sum(  # type: ignore[no-any-return]
        positions[direction],
        axis=2,
    ).T


def unconditional_central_moment(
    data: FDataGrid,
    n: int,
    *,
    domain: DomainRangeLike | None = None,
) -> NDArrayFloat:
    r"""
    Calculate a specified unconditional central moment.

    The unconditional central moments are defined as the unconditional
    moments where the mean is subtracted from each sample before the
    integration. The n-th unconditional central moment is calculated as
    follows, where p is the number of observations:

    .. math::
        f_1(x(t))=\frac{1}{\left(b-a\right)}\int_a^b
        \left(x_1(t) - \mu_1\right)^n dt, \dots,
        f_p(x(t))=\frac{1}{\left(b-a\right)}\int_a^b
        \left(x_p(t) - \mu_p\right)^n dt

        Args:
            data: FDataGrid where we want to calculate
                a particular unconditional central moment.
            n: order of the moment.
            domain: Integration domain. By default, the whole domain is used.

        Returns:
            ndarray of shape (n_dimensions, n_samples) with the values of the
            specified moment.

    Example:
        We will calculate the first unconditional central moment of the
        Canadian Weather data set. In order to simplify the example, we will
        use only the first five samples.
        First we proceed to import the data set.

        >>> from skfda.datasets import fetch_weather
        >>> X = fetch_weather(return_X_y=True)[0]

        Then we call the function with the samples that we want to consider and
        the specified moment order.

        >>> import numpy as np
        >>> from skfda.preprocessing.feature_construction import (
        ...    unconditional_central_moment,
        ... )
        >>> np.around(unconditional_central_moment(X[:5], 1), decimals=2)
        array([[ 0.01,  0.01],
               [ 0.02,  0.01],
               [ 0.02,  0.01],
               [ 0.02,  0.01],
               [ 0.01,  0.01]])

    """
    mean = unconditional_expected_value(
        data,
        lambda x: x,
        domain=domain,
    )

    return unconditional_expected_value(
        data,
        lambda x: np.power(x - mean, n),
        domain=domain,
    )


def unconditional_moment(
    data: Union[FDataBasis, FDataGrid],
    n: int,
    *,
    domain: DomainRangeLike | None = None,
) -> NDArrayFloat:
    r"""
    Calculate a specified unconditional moment.

    The n-th unconditional moments of p real-valued continuous functions
    are calculated as:
    .. math::
        f_1(x(t))=\frac{1}{\left( b-a\right)}\int_a^b \left(x_1(t)\right)^ndt,
        \dots,
        f_p(x(t))=\frac{1}{\left( b-a\right)}\int_a^b  \left(x_p(t)\right)^n dt
        Args:
            data: FDataGrid or FDataBasis where we want to calculate
                a particular unconditional moment.
            n: Order of the moment.
            domain: Integration domain. By default, the whole domain is used.

        Returns:
            ndarray of shape (n_dimensions, n_samples) with the values of the
            specified moment.

    Examples:
        We will calculate the first unconditional moment of the Canadian
        Weather data set. In order to simplify the example, we will use only
        the first five samples.
        First we proceed to import the data set.

        >>> from skfda.datasets import fetch_weather
        >>> X = fetch_weather(return_X_y=True)[0]

        Then we call the function with the samples that we want to consider and
        the specified moment order.

        >>> import numpy as np
        >>> from skfda.preprocessing.feature_construction import (
        ...     unconditional_moment,
        ... )
        >>> np.around(unconditional_moment(X[:5], 1), decimals=2)
        array([[ 4.7 ,  4.03],
               [ 6.16,  3.96],
               [ 5.52,  4.01],
               [ 6.82,  3.44],
               [ 5.25,  3.29]])

    """
    return unconditional_expected_value(
        data,
        lambda x: np.power(x, n),
        domain=domain,
    )


def unconditional_expected_value(
    data: FData,
    function: _UnaryUfunc,
    *,
    domain: DomainRangeLike | None = None,
) -> NDArrayFloat:
    r"""
    Calculate the unconditional expected value of a function.

    Next formula shows for a defined transformation :math: `g(x(t))`
    and p observations, how the unconditional expected values are calculated:

    .. math::
            f_1(x(t))=\frac{1}{\left( b-a\right)}\int_a^b g
            \left(x_1(t)\right)dt,\dots,
            f_p(x(t))=\frac{1}{\left( b-a\right)}\int_a^b g
            \left(x_p(t)\right) dt

    Args:
        data: FDataGrid or FDataBasis where we want to calculate
            the expected value.
        function: function that specifies how the expected value to is
            calculated. It has to be a function of X(t).
        domain: Integration domain. By default, the whole domain is used.

    Returns:
        ndarray of shape (n_dimensions, n_samples) with the values of the
        expectations.

    Examples:
        We will use this function to calculate the logarithmic first moment
        of the first 5 samples of the Berkeley Growth dataset.
        We will start by importing it.

        >>> from skfda.datasets import fetch_growth
        >>> X = fetch_growth(return_X_y=True)[0]

        We will define a function that calculates the inverse first moment.

        >>> import numpy as np
        >>> f = lambda x: np.power(np.log(x), 1)

        Then we call the function with the dataset and the function.

        >>> from skfda.preprocessing.feature_construction import (
        ...     unconditional_expected_value,
        ... )
        >>> np.around(unconditional_expected_value(X[:5], f), decimals=2)
            array([[ 4.96],
                   [ 4.88],
                   [ 4.85],
                   [ 4.9 ],
                   [ 4.84]])

    """
    return _average_function_ufunc(data, ufunc=function, domain=domain)



from __future__ import annotations

import warnings
from typing import Any, Sequence, Tuple, Type, TypeVar

import numpy as np
from numpy import polyint, polymul, polyval
from scipy.interpolate import BSpline as SciBSpline, PPoly

from ...typing._base import DomainRangeLike
from ...typing._numpy import NDArrayFloat
from ._basis import Basis

T = TypeVar("T", bound="BSplineBasis")


class BSplineBasis(Basis):
    r"""BSpline basis.

    BSpline basis elements are defined recursively as:

    .. math::
        B_{i, 1}(x) = 1 \quad \text{if } t_i \le x < t_{i+1},
        \quad 0 \text{ otherwise}

    .. math::
        B_{i, k}(x) = \frac{x - t_i}{t_{i+k} - t_i} B_{i, k-1}(x)
        + \frac{t_{i+k+1} - x}{t_{i+k+1} - t_{i+1}} B_{i+1, k-1}(x)

    Where k indicates the order of the spline.

    Implementation details: In order to allow a discontinuous behaviour at
    the boundaries of the domain it is necessary to placing m knots at the
    boundaries\ :footcite:p:`ramsay+silverman_2005_functionala`. This is
    automatically done so that the user only has to specify a single knot
    at the boundaries.

    Parameters:
        domain_range: A tuple of length 2 containing the initial and
            end values of the interval over which the basis can be evaluated.
        n_basis: Number of functions in the basis.
        order: Order of the splines. One greather than their degree.
        knots: List of knots of the spline functions.

    Examples:
        Constructs specifying number of basis and order.

        >>> bss = BSplineBasis(n_basis=8, order=4)

        If no order is specified defaults to 4 because cubic splines are
        the most used. So the previous example is the same as:

        >>> bss = BSplineBasis(n_basis=8)

        It is also possible to create a BSpline basis specifying the knots.

        >>> bss = BSplineBasis(knots=[0, 0.2, 0.4, 0.6, 0.8, 1])

        Once we create a basis we can evaluate each of its functions at a
        set of points.

        >>> bss = BSplineBasis(n_basis=3, order=3)
        >>> bss([0, 0.5, 1])
        array([[[ 1.  ],
                [ 0.25],
                [ 0.  ]],
               [[ 0.  ],
                [ 0.5 ],
                [ 0.  ]],
               [[ 0.  ],
                [ 0.25],
                [ 1.  ]]])

        And evaluates first derivative

        >>> deriv = bss.derivative()
        >>> deriv([0, 0.5, 1])
        array([[[-2.],
                [-1.],
                [ 0.]],
               [[ 2.],
                [ 0.],
                [-2.]],
               [[ 0.],
                [ 1.],
                [ 2.]]])

    References:
        .. footbibliography::

    """

    def __init__(  # noqa: WPS238
        self,
        domain_range: DomainRangeLike | None = None,
        n_basis: int | None = None,
        order: int = 4,
        knots: Sequence[float] | None = None,
    ) -> None:
        """Bspline basis constructor."""
        from ...misc.validation import validate_domain_range

        if domain_range is not None:
            domain_range = validate_domain_range(domain_range)

            if len(domain_range) != 1:
                raise ValueError("Domain range should be unidimensional.")

            domain_range = domain_range[0]

        # Knots default to equally space points in the domain_range
        if knots is not None:
            knots = tuple(knots)
            knots = sorted(knots)
            if domain_range is None:
                domain_range = (knots[0], knots[-1])
            elif domain_range[0] != knots[0] or domain_range[1] != knots[-1]:
                raise ValueError(
                    "The ends of the knots must be the same "
                    "as the domain_range.",
                )

        # n_basis default to number of knots + order of the splines - 2
        if n_basis is None:
            if knots is None:
                raise ValueError(
                    "Must provide either a list of knots or the"
                    "number of basis.",
                )
            n_basis = len(knots) + order - 2

        if n_basis < order:
            raise ValueError(
                f"The number of basis ({n_basis}) should not be smaller "
                f"than the order of the bspline ({order}).",
            )

        self._order = order
        self._knots = None if knots is None else tuple(knots)
        super().__init__(domain_range=domain_range, n_basis=n_basis)

        # Checks
        if self.n_basis != self.order + len(self.knots) - 2:
            raise ValueError(
                f"The number of basis ({self.n_basis}) has to "
                f"equal the order ({self.order}) plus the "
                f"number of knots ({len(self.knots)}) minus 2.",
            )

    @property
    def knots(self) -> Tuple[float, ...]:
        if self._knots is None:
            return tuple(
                np.linspace(
                    *self.domain_range[0],
                    self.n_basis - self.order + 2,
                ),
            )

        return self._knots

    @property
    def order(self) -> int:
        return self._order

    def _evaluation_knots(self) -> Tuple[float, ...]:
        """
        Get the knots adding m knots to the boundary.

        This needs to be done in order to allow a discontinuous behaviour
        at the boundaries of the domain (see references).

        """
        return tuple(
            (self.knots[0],) * (self.order - 1)
            + self.knots
            + (self.knots[-1],) * (self.order - 1),
        )

    def _evaluate(self, eval_points: NDArrayFloat) -> NDArrayFloat:

        # Input is scalar
        eval_points = eval_points[..., 0]

        return self._to_scipy_bspline(  # type: ignore[no-any-return]
            np.eye(self.n_basis),
        )(eval_points).T

    def _derivative_basis_and_coefs(
        self: T,
        coefs: NDArrayFloat,
        order: int = 1,
    ) -> Tuple[T, NDArrayFloat]:

        if order >= self.order:
            return (
                type(self)(n_basis=1, domain_range=self.domain_range, order=1),
                np.zeros((len(coefs), 1)),
            )

        deriv_splines = self._to_scipy_bspline(coefs).derivative(order)
        return self._from_scipy_bspline(deriv_splines)

    def rescale(  # noqa: D102
        self: T,
        domain_range: DomainRangeLike | None = None,
    ) -> T:
        from ...misc.validation import validate_domain_range

        knots = np.array(self.knots, dtype=np.dtype("float"))

        if domain_range is not None:  # Rescales the knots
            domain_range = validate_domain_range(domain_range)[0]
            knots -= knots[0]
            knots *= (domain_range[1] - domain_range[0]) / (
                self.knots[-1] - self.knots[0]
            )
            knots += domain_range[0]

            # Fix possible round error
            knots[0] = domain_range[0]
            knots[-1] = domain_range[1]

        else:
            # TODO: Allow multiple dimensions
            domain_range = self.domain_range[0]

        return type(self)(domain_range, self.n_basis, self.order, tuple(knots))

    def __repr__(self) -> str:
        """Representation of a BSpline basis."""
        return (
            f"{self.__class__.__name__}(domain_range={self.domain_range}, "
            f"n_basis={self.n_basis}, order={self.order}, "
            f"knots={self.knots})"
        )

    def _gram_matrix(self) -> NDArrayFloat:
        # Places m knots at the boundaries
        knots = self._evaluation_knots()

        # c is used the select which spline the function
        # PPoly.from_spline below computes
        c = np.zeros(len(knots))

        # Initialise empty list to store the piecewise polynomials
        ppoly_lst = []

        no_0_intervals = np.where(np.diff(knots) > 0)[0]

        # For each basis gets its piecewise polynomial representation
        for n in range(self.n_basis):

            # Write a 1 in c in the position of the spline
            # transformed in each iteration
            c[n] = 1

            # Gets the piecewise polynomial representation and gets
            # only the positions for no zero length intervals
            # This polynomial are defined relatively to the knots
            # meaning that the column i corresponds to the ith knot.
            # Let the ith knot be a
            # Then f(x) = pp(x - a)
            pp = PPoly.from_spline((knots, c, self.order - 1))
            pp_coefs = pp.c[:, no_0_intervals]

            # We have the coefficients for each interval in coordinates
            # (x - a), so we will need to subtract a when computing the
            # definite integral
            ppoly_lst.append(pp_coefs)
            c[n] = 0

        # Now for each pair of basis computes the inner product after
        # applying the linear differential operator
        matrix = np.zeros((self.n_basis, self.n_basis))

        for interval, _ in enumerate(no_0_intervals):
            for i in range(self.n_basis):
                poly_i = np.trim_zeros(
                    ppoly_lst[i][:, interval],
                    "f",
                )
                # Indefinite integral
                square = polymul(poly_i, poly_i)
                integral = polyint(square)

                # Definite integral
                matrix[i, i] += np.diff(
                    polyval(
                        integral,
                        np.array(self.knots[interval:interval + 2])
                        - self.knots[interval],
                    ),
                )[0]

                # The Gram matrix is banded, so not all intervals are used
                for j in range(i + 1, min(i + self.order, self.n_basis)):
                    poly_j = np.trim_zeros(ppoly_lst[j][:, interval], "f")

                    # Indefinite integral
                    integral = polyint(polymul(poly_i, poly_j))

                    # Definite integral
                    matrix[i, j] += np.diff(
                        polyval(
                            integral,
                            np.array(self.knots[interval:interval + 2])
                            - self.knots[interval],
                        ),
                    )[0]

                    # The matrix is symmetric
                    matrix[j, i] = matrix[i, j]

        return matrix

    def _to_scipy_bspline(self, coefs: NDArrayFloat) -> SciBSpline:

        repeated_initial: NDArrayFloat = np.repeat(
            self.knots[0],
            self.order - 1,
        )
        knots_array = np.asarray(self.knots)
        repeated_final: NDArrayFloat = np.repeat(
            self.knots[-1],
            self.order - 1,
        )

        knots: NDArrayFloat = np.concatenate(
            (
                repeated_initial,
                knots_array,
                repeated_final,
            ),
        )

        return SciBSpline(knots, coefs.T, self.order - 1)

    @classmethod
    def _from_scipy_bspline(
        cls: Type[T],
        bspline: SciBSpline,
    ) -> Tuple[T, NDArrayFloat]:
        order = bspline.k
        knots = bspline.t

        # Remove additional knots at the borders
        if order != 0:
            knots = knots[order:-order]

        coefs = bspline.c
        domain_range = [knots[0], knots[-1]]

        n_basis = len(knots) + order - 1
        coefs = coefs[:n_basis, ...].T

        return cls(domain_range, order=order + 1, knots=knots), coefs

    def __eq__(self, other: Any) -> bool:
        return (
            super().__eq__(other)
            and self.order == other.order
            and self.knots == other.knots
        )

    def __hash__(self) -> int:
        return hash((super().__hash__(), self.order, self.knots))


class BSpline(BSplineBasis):
    r"""BSpline basis.

    BSpline basis elements are defined recursively as:

    .. math::
        B_{i, 1}(x) = 1 \quad \text{if } t_i \le x < t_{i+1},
        \quad 0 \text{ otherwise}

    .. math::
        B_{i, k}(x) = \frac{x - t_i}{t_{i+k} - t_i} B_{i, k-1}(x)
        + \frac{t_{i+k+1} - x}{t_{i+k+1} - t_{i+1}} B_{i+1, k-1}(x)

    Where k indicates the order of the spline.

    .. deprecated:: 0.8
        Use BSplineBasis instead.

    Implementation details: In order to allow a discontinuous behaviour at
    the boundaries of the domain it is necessary to placing m knots at the
    boundaries\ :footcite:p:`ramsay+silverman_2005_functionala`. This is
    automatically done so that the user only has to specify a single knot
    at the boundaries.

    Parameters:
        domain_range: A tuple of length 2 containing the initial and
            end values of the interval over which the basis can be evaluated.
        n_basis: Number of functions in the basis.
        order: Order of the splines. One greather than their degree.
        knots: List of knots of the spline functions.

    Examples:
        Constructs specifying number of basis and order.

        >>> bss = BSpline(n_basis=8, order=4)

        If no order is specified defaults to 4 because cubic splines are
        the most used. So the previous example is the same as:

        >>> bss = BSpline(n_basis=8)

        It is also possible to create a BSpline basis specifying the knots.

        >>> bss = BSpline(knots=[0, 0.2, 0.4, 0.6, 0.8, 1])

        Once we create a basis we can evaluate each of its functions at a
        set of points.

        >>> bss = BSpline(n_basis=3, order=3)
        >>> bss([0, 0.5, 1])
        array([[[ 1.  ],
                [ 0.25],
                [ 0.  ]],
               [[ 0.  ],
                [ 0.5 ],
                [ 0.  ]],
               [[ 0.  ],
                [ 0.25],
                [ 1.  ]]])

        And evaluates first derivative

        >>> deriv = bss.derivative()
        >>> deriv([0, 0.5, 1])
        array([[[-2.],
                [-1.],
                [ 0.]],
               [[ 2.],
                [ 0.],
                [-2.]],
               [[ 0.],
                [ 1.],
                [ 2.]]])

    References:
        .. footbibliography::

    """

    def __init__(  # noqa: WPS238
        self,
        domain_range: DomainRangeLike | None = None,
        n_basis: int | None = None,
        order: int = 4,
        knots: Sequence[float] | None = None,
    ) -> None:
        super().__init__(
            domain_range=domain_range,
            n_basis=n_basis,
            order=order,
            knots=knots,
        )
        warnings.warn(
            "The BSpline class is deprecated. Use BSplineBasis instead.",
            DeprecationWarning,
            stacklevel=2,
        )



"""Parametric Plot Module.

This module contains the functionality in charge of plotting
two different functions as coordinates, this can be done giving
one FData, with domain 1 and codomain 2, or giving two FData, both
of them with domain 1 and codomain 1.
"""
from __future__ import annotations

from typing import Dict, Sequence, TypeVar

import numpy as np
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.figure import Figure

from ...representation import FData
from ._baseplot import BasePlot
from ._utils import ColorLike
from .representation import Indexable, _get_color_info

K = TypeVar('K', contravariant=True)


class ParametricPlot(BasePlot):
    """
    Parametric Plot visualization.

    This class contains the functionality in charge of plotting
    two different functions as coordinates, this can be done giving
    one FData, with domain 1 and codomain 2, or giving two FData, both
    of them with domain 1 and codomain 1.

    Args:
        fdata1: functional data set that we will use for the graph. If it has
            a dim_codomain = 1, the fdata2 will be needed.
        fdata2: optional functional data set, that will be needed if the fdata1
            has dim_codomain = 1.
        chart: figure over with the graphs are plotted or axis over
            where the graphs are plotted. If None and ax is also
            None, the figure is initialized.
        fig: figure over with the graphs are plotted in case ax is not
            specified. If None and ax is also None, the figure is
            initialized.
        ax: axis where the graphs are plotted. If None, see param fig.
    """

    def __init__(
        self,
        fdata1: FData,
        fdata2: FData | None = None,
        chart: Figure | Axes | None = None,
        *,
        fig: Figure | None = None,
        axes: Axes | None = None,
        group: Sequence[K] | None = None,
        group_colors: Indexable[K, ColorLike] | None = None,
        group_names: Indexable[K, str] | None = None,
        legend: bool = False,
    ) -> None:
        BasePlot.__init__(
            self,
            chart,
            fig=fig,
            axes=axes,
        )
        self.fdata1 = fdata1
        self.fdata2 = fdata2

        if self.fdata2 is not None:
            self.fd_final = self.fdata1.concatenate(
                self.fdata2, as_coordinates=True,
            )
        else:
            self.fd_final = self.fdata1

        self.group = group
        self.group_names = group_names
        self.group_colors = group_colors
        self.legend = legend

    @property
    def n_samples(self) -> int:
        return self.fd_final.n_samples

    def _plot(
        self,
        fig: Figure,
        axes: Axes,
    ) -> None:

        self.artists = np.zeros((self.n_samples, 1), dtype=Artist)

        sample_colors, patches = _get_color_info(
            self.fd_final,
            self.group,
            self.group_names,
            self.group_colors,
            self.legend,
        )

        color_dict: Dict[str, ColorLike | None] = {}

        if (
            self.fd_final.dim_domain == 1
            and self.fd_final.dim_codomain == 2
        ):
            ax = axes[0]

            for i in range(self.fd_final.n_samples):

                if sample_colors is not None:
                    color_dict["color"] = sample_colors[i]

                self.artists[i, 0] = ax.plot(
                    self.fd_final.data_matrix[i][:, 0].tolist(),
                    self.fd_final.data_matrix[i][:, 1].tolist(),
                    **color_dict,
                )[0]

        else:
            raise ValueError(
                "Error in data arguments,",
                "codomain or domain is not correct.",
            )

        if self.fd_final.dataset_name is not None:
            fig.suptitle(self.fd_final.dataset_name)

        if self.fd_final.coordinate_names[0] is None:
            ax.set_xlabel("Function 1")
        else:
            ax.set_xlabel(self.fd_final.coordinate_names[0])

        if self.fd_final.coordinate_names[1] is None:
            ax.set_ylabel("Function 2")
        else:
            ax.set_ylabel(self.fd_final.coordinate_names[1])



"""Neighbors outlier detection methods."""
from __future__ import annotations

from typing import Any, TypeVar, Union, overload

from sklearn.base import OutlierMixin
from sklearn.neighbors import LocalOutlierFactor as _LocalOutlierFactor
from typing_extensions import Literal

from ..._utils._neighbors_base import AlgorithmType, KNeighborsMixin
from ...misc.metrics import PairwiseMetric, l2_distance
from ...representation import FData
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat, NDArrayInt

SelfType = TypeVar("SelfType", bound="LocalOutlierFactor[Any]")
InputBound = Union[NDArrayFloat, FData]
Input = TypeVar("Input", contravariant=True, bound=InputBound)


class LocalOutlierFactor(
    KNeighborsMixin[Input, None],
    OutlierMixin,  # type: ignore[misc]
):
    r"""
    Unsupervised Outlier Detection.

    Unsupervised Outlier Detection using Local Outlier Factor (LOF).

    The anomaly score of each sample is called Local Outlier Factor.
    It measures the local deviation of density of a given sample with
    respect to its neighbors.

    It is local in that the anomaly score depends on how isolated the object
    is with respect to the surrounding neighborhood.

    More precisely, locality is given by k-nearest neighbors, whose distance
    is used to estimate the local density.

    By comparing the local density of a sample to the local densities of
    its neighbors, one can identify samples that have a substantially lower
    density than their neighbors. These are considered outliers.

    Parameters:
        n_neighbors: Number of neighbors to use by default for
            :meth:`kneighbors` queries.
            If n_neighbors is larger than the number of samples provided,
            all samples will be used.
        algorithm: Algorithm used to compute the nearest neighbors:

            - 'ball_tree' will use :class:`BallTree`
            - 'kd_tree' will use :class:`KDTree`
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
              based on the values passed to :meth:`fit` method.

        leaf_size: Leaf size passed to :class:`BallTree` or :class:`KDTree`.
            This can affect the speed of the construction and query, as well as
            the memory required to store the tree. The optimal value depends on
            the nature of the problem.
        metric: The distance metric to use for the tree.  The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        contamination: The amount of contamination of the data set, i.e. the
            proportion of outliers in the data set. When fitting this is used
            to define the threshold on the decision function. If "auto", the
            decision function threshold is determined as in the original paper
            [BKNS2000]_.
        novelty: By default, LocalOutlierFactor is only meant to be used for
            outlier detection (novelty=False). Set novelty to True if you want
            to use LocalOutlierFactor for novelty detection. In this case be
            aware that that you should only use predict, decision_function and
            score_samples on new unseen data and not on the training set.
        n_jobs: The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors. See :term:`Glossary <n_jobs>`
            for more details.
            Affects only :meth:`kneighbors` and :meth:`kneighbors_graph`
            methods.

    Attributes:
        negative_outlier_factor\_ : The opposite LOF of the training samples.
            The higher, the more normal. Inliers tend to have a LOF score close
            to 1 (``negative_outlier_factor_`` close to -1), while outliers
            tend to have a larger LOF score.
            The local outlier factor (LOF) of a sample captures its
            supposed 'degree of abnormality'.
            It is the average of the ratio of the local reachability density of
            a sample and those of its k-nearest neighbors.
        n_neighbors\_ : The actual number of neighbors used for
            :meth:`kneighbors` queries.
        offset\_ : Offset used to obtain binary labels from the raw scores.
            Observations having a negative_outlier_factor smaller than
            `offset_` are detected as abnormal.
            The offset is set to -1.5 (inliers score around -1), except when a
            contamination parameter different than "auto" is provided. In that
            case, the offset is defined in such a way we obtain the expected
            number of outliers in training.

    Examples:
        **Local Outlier Factor (LOF) for outlier detection**.

        >>> from skfda.exploratory.outliers import LocalOutlierFactor

        Creation of simulated dataset with 2 outliers to be used with LOF.

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd_clean = make_sinusoidal_process(n_samples=25, error_std=0,
        ...                                    phase_std=0.1, random_state=0)
        >>> fd_outliers = make_sinusoidal_process(
        ...     n_samples=2, error_std=0, phase_mean=0.5, random_state=5)
        >>> fd = fd_outliers.concatenate(fd_clean) # Dataset with 2 outliers

        Detection of outliers with LOF.

        >>> lof = LocalOutlierFactor()
        >>> is_outlier = lof.fit_predict(fd)
        >>> is_outlier # -1 for anomalies/outliers and +1 for inliers
        array([-1, -1,  1,  1,  1,  1,  1,  1, ...,  1,  1,  1,  1])

        The negative outlier factor stored.

        >>> lof.negative_outlier_factor_.round(2)
        array([-7.11, -1.54, -1.  , -0.99, ..., -0.97,  -1. ,  -0.99])

        **Novelty detection with LOF**.

        Creation of a dataset without outliers.

        >>> fd_train = make_sinusoidal_process(
        ...     n_samples=25,
        ...     error_std=0,
        ...     phase_std=0.1,
        ...     random_state=9,
        ... )

        Fit of LOF using the dataset without outliers.

        >>> lof = LocalOutlierFactor(novelty=True)
        >>> lof.fit(fd_train)
        LocalOutlierFactor(...novelty=True)

        Detection of annomalies for new samples.

        >>> lof.predict(fd) # Predict with samples not used in fit
        array([-1, -1,  1,  1,  1,  1,  1,  1, ...,  1,  1,  1,  1])


    References:
        .. [BKNS2000] Breunig, M. M., Kriegel, H. P., Ng, R. T., & Sander,
           J. (2000, May). LOF: identifying density-based local outliers.
           In ACM sigmod record.

    Notes:
        This estimator wraps the scikit-learn class
        :class:`~sklearn.neighbors.LocalOutlierFactor` employing functional
        metrics and data instead of the multivariate ones.

    See also:
        :class:`~skfda.ml.classification.KNeighborsClassifier`
        :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroids`
        :class:`~skfda.ml.regression.KNeighborsRegressor`
        :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
        :class:`~skfda.ml.clustering.NearestNeighbors`
    """

    @overload
    def __init__(
        self: LocalOutlierFactor[NDArrayFloat],
        *,
        n_neighbors: int = 20,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"],
        contamination: float | Literal["auto"] = "auto",
        novelty: bool = False,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: LocalOutlierFactor[InputBound],
        *,
        n_neighbors: int = 20,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        contamination: float | Literal["auto"] = "auto",
        novelty: bool = False,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        n_neighbors: int = 20,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        contamination: float | Literal["auto"] = "auto",
        novelty: bool = False,
        n_jobs: int | None = None,
    ) -> None:
        pass

    def __init__(
        self,
        *,
        n_neighbors: int = 20,
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        contamination: float | Literal["auto"] = "auto",
        novelty: bool = False,
        n_jobs: int | None = None,
    ) -> None:
        super().__init__(
            n_neighbors=n_neighbors,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            n_jobs=n_jobs,
        )
        self.contamination = contamination
        self.novelty = novelty

    def _init_estimator(self) -> _LocalOutlierFactor:
        return _LocalOutlierFactor(
            n_neighbors=self.n_neighbors,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric="precomputed",
            contamination=self.contamination,
            novelty=self.novelty,
            n_jobs=self.n_jobs,
        )

    def _store_fit_data(self) -> None:
        """Store the parameters created during the fit."""
        self.negative_outlier_factor_ = (
            self._estimator.negative_outlier_factor_
        )
        self.n_neighbors_ = self._estimator.n_neighbors_
        self.offset_ = self._estimator.offset_

    def fit(  # noqa: D102
        self: SelfType,
        X: Input,
        y: None = None,
    ) -> SelfType:
        self._fit(X, y, fit_with_zeros=False)
        self._store_fit_data()

        return self

    def predict(
        self,
        X: Input,
    ) -> NDArrayInt:
        """
        Predict the labels (1 inlier, -1 outlier) of X according to LOF.

        This method allows to generalize prediction to *new observations* (not
        in the training set). Only available for novelty detection (when
        novelty is set to True).

        If X is None, returns the same as fit_predict(X_train).

        Parameters:
            X: FDataGrid containing the query sample or samples to compute the
                Local Outlier Factor w.r.t. to the training samples or array
                with the distances to the training samples if
                metric='precomputed'.

        Returns:
            Returns -1 for anomalies/outliers and +1 for inliers.

        """
        self._check_is_fitted()
        X_dist = self._X_to_distances(X)

        return self._estimator.predict(X_dist)  # type: ignore[no-any-return]

    def fit_predict(
        self,
        X: Input,
        y: None = None,
    ) -> NDArrayInt:
        """
        Fits the model to the training set X and returns the labels.

        Label is 1 for an inlier and -1 for an outlier according to the LOF
        score and the contamination parameter.

        Parameters:
            X: Training data. Samples, or array with shape
                [n_samples, n_samples] if metric='precomputed'.
            y : Ignored. Not used, present for API consistency by convention.

        Returns:
            Returns -1 for anomalies/outliers and 1 for inliers.

        """
        # In this estimator fit_predict cannot be wrapped as fit().predict()

        self._estimator = self._init_estimator()
        metric = self.metric

        if metric == 'precomputed':
            res = self._estimator.fit_predict(X, y)
        else:
            X_dist = PairwiseMetric(metric)(X)
            res = self._estimator.fit_predict(X_dist, y)

        self._store_fit_data()
        self._fitted_with_distances = True

        return res  # type: ignore[no-any-return]

    def decision_function(
        self,
        X: Input,
    ) -> NDArrayFloat:
        """
        Shifted opposite of the Local Outlier Factor of X.

        Bigger is better, i.e. large values correspond to inliers.
        The shift offset allows a zero threshold for being an outlier.
        Only available for novelty detection (when novelty is set to True).
        The argument X is supposed to contain *new data*: if X contains a
        point from training, it considers the later in its own neighborhood.
        Also, the samples in X are not considered in the neighborhood of any
        point.

        Parameters:
            X: Query sample or samples to compute the
                Local Outlier Factor w.r.t. to the training samples.

        Returns:
            The shifted opposite of the Local Outlier Factor of each input
            samples. The lower, the more abnormal. Negative scores represent
            outliers, positive scores represent inliers.

        """
        self._check_is_fitted()
        X_dist = self._X_to_distances(X)

        return (  # type: ignore[no-any-return]
            self._estimator.decision_function(X_dist)
        )

    def score_samples(
        self,
        X: Input,
    ) -> NDArrayFloat:
        """
        Opposite of the Local Outlier Factor of X.

        It is the opposite as bigger is better, i.e. large values correspond
        to inliers.

        Only available for novelty detection (when novelty is set to True).
        The argument X is supposed to contain *new data*: if X contains a
        point from training, it considers the later in its own neighborhood.
        Also, the samples in X are not considered in the neighborhood of any
        point.

        The score_samples on training data is available by considering the
        the ``negative_outlier_factor_`` attribute.

        Parameters:
            X: Query sample or samples to compute the
                Local Outlier Factor w.r.t. to the training samples.

        Returns:
            The opposite of the Local Outlier Factor of each input samples.
            The lower, the more abnormal.

        """
        self._check_is_fitted()
        X_dist = self._X_to_distances(X)

        return (  # type: ignore[no-any-return]
            self._estimator.score_samples(X_dist)
        )



"""BasePlot Module.

This module contains the abstract class of which inherit all
the visualization modules, containing the basic functionality
common to all of them.
"""
from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Sequence, Tuple

import matplotlib.pyplot as plt
from matplotlib.artist import Artist
from matplotlib.axes import Axes
from matplotlib.backend_bases import LocationEvent, MouseEvent
from matplotlib.collections import PathCollection
from matplotlib.colors import ListedColormap
from matplotlib.figure import Figure
from matplotlib.text import Annotation

from ...representation import FData
from ...typing._numpy import NDArrayInt, NDArrayObject
from ._utils import _figure_to_svg, _get_figure_and_axes, _set_figure_layout


class BasePlot(ABC):
    """
    BasePlot class.

    Attributes:
        artists: List of Artist objects corresponding
            to every instance of our plot. They will be used to modify
            the visualization with interactivity and widgets.
        fig: Figure over with the graphs are plotted.
        axes: Sequence of axes where the graphs are plotted.
    """

    @abstractmethod
    def __init__(
        self,
        chart: Figure | Axes | None = None,
        *,
        fig: Figure | None = None,
        axes: Axes | Sequence[Axes] | None = None,
        n_rows: int | None = None,
        n_cols: int | None = None,
        c: NDArrayInt | None = None,
        cmap_bold: ListedColormap = None,
        x_label: str | None = None,
        y_label: str | None = None,
    ) -> None:
        self.artists: NDArrayObject | None = None
        self.chart = chart
        self.fig = fig
        self.axes = axes
        self.n_rows = n_rows
        self.n_cols = n_cols
        self._tag = self._create_annotation()
        self.c = c
        self.cmap_bold = cmap_bold
        self.x_label = x_label
        self.y_label = y_label

    def _plot(
        self,
        fig: Figure,
        axes: Sequence[Axes],
    ) -> None:
        pass

    def plot(
        self,
    ) -> Figure:
        """
        Plot the object and its data.

        Returns:
            Figure: figure object in which the displays and
                widgets will be plotted.
        """
        fig: Figure | None = getattr(self, "fig_", None)
        axes: Sequence[Axes] | None = getattr(self, "axes_", None)

        if fig is None:
            fig, axes = self._set_figure_and_axes(
                self.chart,
                fig=self.fig,
                axes=self.axes,
            )

        assert axes is not None

        if self.x_label is not None:
            axes[0].set_xlabel(self.x_label)
        if self.y_label is not None:
            axes[0].set_ylabel(self.y_label)

        self._plot(fig, axes)

        self._hover_event_id = fig.canvas.mpl_connect(
            'motion_notify_event',
            self.hover,
        )

        return fig

    @property
    def dim(self) -> int:
        """Get the number of dimensions for this plot."""
        return 2

    @property
    def n_subplots(self) -> int:
        """Get the number of subplots that this plot uses."""
        return 1

    @property
    def n_samples(self) -> int | None:
        """Get the number of instances that will be used for interactivity."""
        return None

    def _set_figure_and_axes(
        self,
        chart: Figure | Axes | None = None,
        *,
        fig: Figure | None = None,
        axes: Axes | Sequence[Axes] | None = None,
    ) -> Tuple[Figure, Sequence[Axes]]:
        fig, axes = _get_figure_and_axes(chart, fig, axes)
        fig, axes = _set_figure_layout(
            fig=fig,
            axes=axes,
            dim=self.dim,
            n_axes=self.n_subplots,
            n_rows=self.n_rows,
            n_cols=self.n_cols,
        )

        self.fig_ = fig
        self.axes_ = axes

        return fig, axes

    def _repr_svg_(self) -> str:
        """Automatically represents the object as an svg when calling it."""
        self.fig = self.plot()
        plt.close(self.fig)
        return _figure_to_svg(self.fig)

    def _create_annotation(self) -> Annotation:
        tag = Annotation(
            "",
            xy=(0, 0),
            xytext=(20, 20),
            textcoords="offset points",
            bbox={
                "boxstyle": "round",
                "fc": "w",
            },
            arrowprops={
                "arrowstyle": "->",
            },
            annotation_clip=False,
            clip_on=False,
        )

        tag.get_bbox_patch().set_facecolor(color='khaki')
        intensity = 0.8
        tag.get_bbox_patch().set_alpha(intensity)

        return tag

    def _update_annotation(
        self,
        tag: Annotation,
        *,
        axes: Axes,
        sample_number: int,
        fdata: FData | None,
        position: Tuple[float, float],
    ) -> None:
        """
        Auxiliary method used to update the hovering annotations.

        Method used to update the annotations that appear while
        hovering a scattered point. The annotations indicate
        the index and coordinates of the point hovered.
        Args:
            tag: Annotation to update.
            axes: Axes were the annotation belongs.
            sample_number: Number of the current sample.
        """
        xdata_graph, ydata_graph = position

        tag.xy = (xdata_graph, ydata_graph)

        sample_name = (
            fdata.sample_names[sample_number]
            if fdata is not None
            else None
        )

        sample_descr = f" ({sample_name})" if sample_name is not None else ""

        text = (
            f"{sample_number}{sample_descr}: "
            f"({xdata_graph:.3g}, {ydata_graph:.3g})"
        )
        tag.set_text(text)

        x_axis = axes.get_xlim()
        y_axis = axes.get_ylim()

        label_xpos = -60
        label_ypos = 20
        if (xdata_graph - x_axis[0]) > (x_axis[1] - xdata_graph):
            label_xpos = -80

        if (ydata_graph - y_axis[0]) > (y_axis[1] - ydata_graph):
            label_ypos = -20

        if tag.figure:
            tag.remove()
        tag.figure = None
        axes.add_artist(tag)
        tag.set_transform(axes.transData)
        tag.set_position((label_xpos, label_ypos))

    def _sample_artist_from_event(
        self,
        event: LocationEvent,
    ) -> Tuple[int, FData | None, Artist] | None:
        """Get the number, fdata and artist under a location event."""
        if self.artists is None:
            return None

        try:
            i = self.axes_.index(event.inaxes)
        except ValueError:
            return None

        for j, artist in enumerate(self.artists[:, i]):
            if not isinstance(artist, PathCollection):
                return None

            if artist.contains(event)[0]:
                return j, getattr(self, "fdata", None), artist

        return None

    def hover(self, event: MouseEvent) -> None:
        """
        Activate the annotation when hovering a point.

        Callback method that activates the annotation when hovering
        a specific point in a graph. The annotation is a description
        of the point containing its coordinates.

        Args:
            event: event object containing the artist of the point
                hovered.

        """
        found_artist = self._sample_artist_from_event(event)

        if event.inaxes is not None and found_artist is not None:
            sample_number, fdata, artist = found_artist

            self._update_annotation(
                self._tag,
                axes=event.inaxes,
                sample_number=sample_number,
                fdata=fdata,
                position=artist.get_offsets()[0],
            )
            self._tag.set_visible(True)
            self.fig_.canvas.draw_idle()
        elif self._tag.get_visible():
            self._tag.set_visible(False)
            self.fig_.canvas.draw_idle()



"""Registration of functional data module.

This module contains methods to perform the registration of
functional data, in basis as well in discretized form.
"""
from typing import TYPE_CHECKING

import lazy_loader as lazy

from ..._utils import (
    invert_warping as invert_warping,
    normalize_warping as normalize_warping,
)

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "validation",
    ],
    submod_attrs={
        "_base": [
            "RegistrationTransformer",
            "InductiveRegistrationTransformer",
        ],
        "_fisher_rao": ["ElasticRegistration", "FisherRaoElasticRegistration"],
        "_landmark_registration": [
            "landmark_elastic_registration",
            "landmark_elastic_registration_warping",
            "landmark_registration",
            "landmark_shift",
            "landmark_shift_deltas",
            "landmark_shift_registration",
        ],
        "_lstsq_shift_registration": [
            "LeastSquaresShiftRegistration",
            "ShiftRegistration",
        ],
    },
)

if TYPE_CHECKING:
    from ._base import (
        InductiveRegistrationTransformer as InductiveRegistrationTransformer,
        RegistrationTransformer as RegistrationTransformer,
    )
    from ._fisher_rao import (
        ElasticRegistration as ElasticRegistration,
        FisherRaoElasticRegistration as FisherRaoElasticRegistration,
    )
    from ._landmark_registration import (
        landmark_elastic_registration as landmark_elastic_registration,
        landmark_elastic_registration_warping as landmark_elastic_registration_warping,
        landmark_registration as landmark_registration,
        landmark_shift as landmark_shift,
        landmark_shift_deltas as landmark_shift_deltas,
        landmark_shift_registration as landmark_shift_registration,
    )
    from ._lstsq_shift_registration import (
        LeastSquaresShiftRegistration as LeastSquaresShiftRegistration,
        ShiftRegistration as ShiftRegistration,
    )



"""Functions for working with arrays of functions."""
from __future__ import annotations

import math
from typing import TYPE_CHECKING, Any, Literal, Protocol, TypeVar

from ...misc.validation import validate_domain_range
from .. import nquad_vec

if TYPE_CHECKING:
    from ...typing._numpy import NDArrayFloat
    from ...representation import FData
    from ...typing._base import DomainRangeLike


UfuncMethod = Literal[
    "__call__",
    "reduce",
    "reduceat",
    "accumulate",
    "outer",
    "inner",
]


class _SupportsArrayUFunc(Protocol):
    def __array_ufunc__(
        self,
        ufunc: Any,
        method: UfuncMethod,
        *inputs: Any,
        **kwargs: Any,
    ) -> Any:
        pass


T = TypeVar("T", bound=_SupportsArrayUFunc)


class _UnaryUfunc(Protocol):

    def __call__(self, __arg: T) -> T:  # noqa: WPS112
        pass


def _average_function_ufunc(
    data: FData,
    ufunc: _UnaryUfunc,
    *,
    domain: DomainRangeLike | None = None,
) -> NDArrayFloat:

    if domain is None:
        domain = data.domain_range
    else:
        domain = validate_domain_range(domain)

    lebesgue_measure = math.prod(
        (
            (iterval[1] - iterval[0])
            for iterval in domain
        ),
    )

    try:
        data_eval = ufunc(data)
    except TypeError:

        def integrand(*args: NDArrayFloat) -> NDArrayFloat:  # noqa: WPS430
            f1 = data(args)[:, 0, :]
            return ufunc(f1)

        return nquad_vec(
            integrand,
            domain,
        ) / lebesgue_measure

    else:
        return data_eval.integrate(domain=domain) / lebesgue_measure


def average_function_value(
    data: FData,
    *,
    domain: DomainRangeLike | None = None,
) -> NDArrayFloat:
    r"""
    Calculate the average function value for each function.

    This is the value that, if integrated over the whole domain of each
    function, has the same integral as the function itself.

    .. math::
        \bar{x} = \frac{1}{\text{Vol}(\mathcal{T})}\int_{\mathcal{T}} x(t) dt

    Args:
        data: Functions where we want to calculate the expected value.
        domain: Integration domain. By default, the whole domain is used.

    Returns:
        ndarray of shape (n_dimensions, n_samples) with the values of the
        expectations.

    See also:
        `Entry on Wikipedia
        <https://en.wikipedia.org/wiki/Mean_of_a_function>`_

    """
    return _average_function_ufunc(data, ufunc=lambda x: x, domain=domain)



"""Tests for Recursive Maxima Hunting (RMH)."""
import unittest

import numpy as np
import pytest

import skfda
from skfda.datasets import make_gaussian_process
from skfda.exploratory.stats.covariance import ParametricGaussianCovariance
from skfda.misc.covariances import Exponential
from skfda.preprocessing.dim_reduction import variable_selection as vs


class TestRMH(unittest.TestCase):
    """Tests for RMH."""

    def test_gaussian_homoscedastic(self) -> None:
        """
        Test the case for which RMH is optimal.

        It tests a case with two homoscedastic Brownian processes where the
        difference of means is piecewise linear.

        In this case RMH should return the points where the linear parts
        join.

        """
        n_samples = 10000
        n_features = 100

        def mean_1(  # noqa: WPS430
            t: np.typing.NDArray[np.float_],
        ) -> np.typing.NDArray[np.float_]:

            return (  # type: ignore[no-any-return]
                np.abs(t - 0.25)
                - 2 * np.abs(t - 0.5)
                + np.abs(t - 0.75)
            )

        X_0 = make_gaussian_process(
            n_samples=n_samples // 2,
            n_features=n_features,
            random_state=0,
        )
        X_1 = make_gaussian_process(
            n_samples=n_samples // 2,
            n_features=n_features,
            mean=mean_1,
            random_state=1,
        )
        X = skfda.concatenate((X_0, X_1))

        y = np.zeros(n_samples)
        y[n_samples // 2:] = 1

        correction = vs.recursive_maxima_hunting.GaussianSampleCorrection()
        stopping_condition = vs.recursive_maxima_hunting.ScoreThresholdStop(
            threshold=0.05,
        )

        rmh = vs.RecursiveMaximaHunting(
            correction=correction,
            stopping_condition=stopping_condition,
        )
        rmh.fit(X, y)
        point_mask = rmh.get_support()
        points = X.grid_points[0][point_mask]
        np.testing.assert_allclose(points, [0.25, 0.5, 0.75], rtol=1e-1)

    @pytest.mark.filterwarnings(
        'ignore::sklearn.exceptions.ConvergenceWarning'
    )
    def test_fit_exponential(self) -> None:
        """
        Test the case for which RMH is optimal.

        It tests a case with two homoscedastic Brownian processes where the
        difference of means is piecewise linear.

        In this case RMH should return the points where the linear parts
        join.

        """
        n_samples = 1000
        n_features = 100

        def mean_1(  # noqa: WPS430
            t: np.typing.NDArray[np.float_],
        ) -> np.typing.NDArray[np.float_]:

            return (  # type: ignore[no-any-return]
                np.abs(t - 0.25)
                - 2 * np.abs(t - 0.5)
                + np.abs(t - 0.75)
            )

        X_0 = make_gaussian_process(
            n_samples=n_samples // 2,
            n_features=n_features,
            cov=Exponential(length_scale=2),
            random_state=0,
        )
        X_1 = make_gaussian_process(
            n_samples=n_samples // 2,
            n_features=n_features,
            mean=mean_1,
            cov=Exponential(length_scale=2),
            random_state=1,
        )
        X = skfda.concatenate((X_0, X_1))

        y = np.zeros(n_samples)
        y[n_samples // 2:] = 1

        correction = vs.recursive_maxima_hunting.GaussianSampleCorrection(
            cov_estimator=ParametricGaussianCovariance(
                cov=Exponential(),
            ),
        )
        stopping_condition = vs.recursive_maxima_hunting.ScoreThresholdStop(
            threshold=0.05,
        )

        rmh = vs.RecursiveMaximaHunting(
            correction=correction,
            stopping_condition=stopping_condition,
        )
        rmh.fit(X, y)
        point_mask = rmh.get_support()
        points = X.grid_points[0][point_mask]
        np.testing.assert_allclose(points, [0.25, 0.5, 0.75], rtol=1e-1)


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations

import abc
from typing import Callable, Sequence, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import sklearn.gaussian_process.kernels as sklearn_kern
from matplotlib.figure import Figure
from scipy.special import gamma, kv

from ..representation import FData, FDataBasis, FDataGrid
from ..representation.basis import TensorBasis
from ..typing._numpy import ArrayLike, NDArrayFloat


def _squared_norms(x: NDArrayFloat, y: NDArrayFloat) -> NDArrayFloat:
    return (  # type: ignore[no-any-return]
        (x[:, np.newaxis, :] - y[np.newaxis, :, :]) ** 2
    ).sum(2)


CovarianceLike = Union[
    float,
    NDArrayFloat,
    Callable[[ArrayLike, ArrayLike], NDArrayFloat],
]


def _transform_to_2d(t: ArrayLike) -> NDArrayFloat:
    """Transform 1d arrays in column vectors."""
    t = np.asfarray(t)

    dim = len(t.shape)
    assert dim <= 2

    if dim < 2:
        t = np.atleast_2d(t).T

    return t


def _execute_covariance(
    covariance: CovarianceLike,
    x: ArrayLike,
    y: ArrayLike,
) -> NDArrayFloat:
    """Execute a covariance function."""
    x = _transform_to_2d(x)
    y = _transform_to_2d(y)

    if isinstance(covariance, (int, float)):
        return np.array(covariance)
    else:
        if callable(covariance):
            result = covariance(x, y)
        elif isinstance(covariance, np.ndarray):
            result = covariance
        else:
            # GPy kernel
            result = covariance.K(x, y)

        assert result.shape[0] == len(x)
        assert result.shape[1] == len(y)
        return result


class Covariance(abc.ABC):
    """Abstract class for covariance functions."""

    _parameters_str: Sequence[Tuple[str, str]]
    _latex_formula: str

    @abc.abstractmethod
    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        pass

    def heatmap(self, limits: Tuple[float, float] = (-1, 1)) -> Figure:
        """Return a heatmap plot of the covariance function."""
        from ..exploratory.visualization._utils import _create_figure

        x = np.linspace(*limits, 1000)

        cov_matrix = self(x, x)

        fig = _create_figure()
        ax = fig.add_subplot(1, 1, 1)
        ax.imshow(
            cov_matrix,
            extent=[limits[0], limits[1], limits[1], limits[0]],
        )
        ax.set_title(f"Covariance function in [{limits[0]}, {limits[1]}]")

        return fig

    def _sample_trajectories_plot(self) -> Figure:
        from ..datasets import make_gaussian_process

        fd = make_gaussian_process(
            start=-1,
            n_samples=10,
            cov=self,
            random_state=0,
        )
        fig = fd.plot()
        fig.axes[0].set_title("Sample trajectories")
        return fig

    def __repr__(self) -> str:

        params_str = ', '.join(
            f'{n}={getattr(self, n)}' for n, _ in self._parameters_str
        )

        return (
            f"{self.__module__}.{type(self).__qualname__}("
            f"{params_str}"
            f")"
        )

    def _latex_content(self) -> str:
        params_str = ''.join(
            fr'{l} &= {getattr(self, n)} \\' for n, l in self._parameters_str
        )

        return (
            fr"{self._latex_formula} \\"
            r"\text{where:}"
            r"\begin{aligned}"
            fr"\qquad{params_str}"
            r"\end{aligned}"
        )

    def _repr_latex_(self) -> str:
        return fr"\(\displaystyle {self._latex_content()}\)"

    def _repr_html_(self) -> str:
        from ..exploratory.visualization._utils import _figure_to_svg

        fig = self.heatmap()
        heatmap = _figure_to_svg(fig)
        plt.close(fig)

        fig = self._sample_trajectories_plot()
        sample_trajectories = _figure_to_svg(fig)
        plt.close(fig)

        row_style = ''

        def column_style(percent: float, margin_top: str = "0") -> str:
            return (
                f'style="display: inline-block; '
                f'margin:0; '
                f'margin-top: {margin_top}; '
                f'width:{percent}%; '
                f'height:auto;'
                f'vertical-align: middle"'
            )

        return fr"""
        <div {row_style}>
            <div {column_style(100, margin_top='25px')}>
            \[{self._latex_content()}\]
            </div>
        </div>
        <div {row_style}>
            <div {column_style(48)}>
            {sample_trajectories}
            </div>
            <div {column_style(48)}>
            {heatmap}
            </div>
        </div>
        """

    def to_sklearn(self) -> sklearn_kern.Kernel:
        """Convert it to a sklearn kernel, if there is one."""
        raise NotImplementedError(
            f"{type(self).__name__} covariance not "
            f"implemented in scikit-learn",
        )


class Brownian(Covariance):
    r"""
    Brownian covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \frac{|x - \mathcal{O}| + |x' - \mathcal{O}|
        - |x - x'|}{2}

    where :math:`\sigma^2` is the variance at distance 1 from
    :math:`\mathcal{O}` and :math:`\mathcal{O}` is the origin point.
    If :math:`\mathcal{O} = 0` (the default) and we only
    consider positive values, the formula can be simplified as

    .. math::
        K(x, y) = \sigma^2 \min(x, y).

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Brownian
        import matplotlib.pyplot as plt

        Brownian().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Brownian
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Brownian(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Brownian

        Brownian()

    """
    _latex_formula = (
        r"K(x, x') = \sigma^2 \frac{|x - \mathcal{O}| + "
        r"|x' - \mathcal{O}| - |x - x'|}{2}"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("origin", r"\mathcal{O}"),
    ]

    def __init__(self, *, variance: float = 1, origin: float = 0) -> None:
        self.variance = variance
        self.origin = origin

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x) - self.origin
        y = _transform_to_2d(y) - self.origin

        sum_norms = np.add.outer(
            np.linalg.norm(x, axis=-1),
            np.linalg.norm(y, axis=-1),
        )
        norm_sub = np.linalg.norm(
            x[:, np.newaxis, :] - y[np.newaxis, :, :],
            axis=-1,
        )

        return (  # type: ignore[no-any-return]
            self.variance * (sum_norms - norm_sub) / 2
        )


class Linear(Covariance):
    r"""
    Linear covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 (x^T x' + c)

    where :math:`\sigma^2` is the scale of the variance and
    :math:`c` is the intercept.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Linear
        import matplotlib.pyplot as plt

        Linear().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Linear
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Linear(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Linear

        Linear()

    """

    _latex_formula = r"K(x, x') = \sigma^2 (x^T x' + c)"

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("intercept", "c"),
    ]

    def __init__(self, *, variance: float = 1, intercept: float = 0) -> None:
        self.variance = variance
        self.intercept = intercept

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        return self.variance * (x @ y.T + self.intercept)

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * (sklearn_kern.DotProduct(0) + self.intercept)
        )


class Polynomial(Covariance):
    r"""
    Polynomial covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 (\alpha x^T x' + c)^d

    where :math:`\sigma^2` is the scale of the variance,
    :math:`\alpha` is the slope, :math:`d` the degree of the
    polynomial and :math:`c` is the intercept.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Polynomial
        import matplotlib.pyplot as plt

        Polynomial().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Polynomial
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Polynomial(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Polynomial

        Polynomial()

    """

    _latex_formula = r"K(x, x') = \sigma^2 (\alpha x^T x' + c)^d"

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("intercept", "c"),
        ("slope", r"\alpha"),
        ("degree", "d"),
    ]

    def __init__(
        self,
        *,
        variance: float = 1,
        intercept: float = 0,
        slope: float = 1,
        degree: float = 2,
    ) -> None:
        self.variance = variance
        self.intercept = intercept
        self.slope = slope
        self.degree = degree

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        return (
            self.variance
            * (self.slope * x @ y.T + self.intercept) ** self.degree
        )

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * (self.slope * sklearn_kern.DotProduct(0) + self.intercept)
            ** self.degree
        )


class Gaussian(Covariance):
    r"""
    Gaussian covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \exp\left(-\frac{||x - x'||^2}{2l^2}\right)

    where :math:`\sigma^2` is the variance and :math:`l` is the length scale.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Gaussian
        import matplotlib.pyplot as plt

        Gaussian().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Gaussian
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Gaussian(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Gaussian

        Gaussian()

    """

    _latex_formula = (
        r"K(x, x') = \sigma^2 \exp\left(-\frac{\|x - x'\|^2}{2l^2}"
        r"\right)"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("length_scale", "l"),
    ]

    def __init__(self, *, variance: float = 1, length_scale: float = 1):
        self.variance = variance
        self.length_scale = length_scale

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        x_y = _squared_norms(x, y)

        return self.variance * np.exp(-x_y / (2 * self.length_scale ** 2))

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance * sklearn_kern.RBF(length_scale=self.length_scale)
        )


class Exponential(Covariance):
    r"""
    Exponential covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \exp\left(-\frac{\|x - x'\|}{l}\right)

    where :math:`\sigma^2` is the variance and :math:`l` is the length scale.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Exponential
        import matplotlib.pyplot as plt

        Exponential().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Exponential
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Exponential(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Exponential

        Exponential()

    """

    _latex_formula = (
        r"K(x, x') = \sigma^2 \exp\left(-\frac{||x - x'||}{l}"
        r"\right)"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("length_scale", "l"),
    ]

    def __init__(
        self,
        *,
        variance: float = 1,
        length_scale: float = 1,
    ) -> None:
        self.variance = variance
        self.length_scale = length_scale

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        x_y = _squared_norms(x, y)
        return self.variance * np.exp(-np.sqrt(x_y) / (self.length_scale))

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * sklearn_kern.Matern(length_scale=self.length_scale, nu=0.5)
        )


class WhiteNoise(Covariance):
    r"""
    Gaussian covariance function.

    The covariance function is

    .. math::
        K(x, x')= \begin{cases}
                    \sigma^2, \quad x = x' \\
                    0, \quad x \neq x'\\
                  \end{cases}

    where :math:`\sigma^2` is the variance.

    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import WhiteNoise
        import matplotlib.pyplot as plt

        WhiteNoise().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import WhiteNoise
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=WhiteNoise(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import WhiteNoise

        WhiteNoise()

    """

    _latex_formula = (
        r"K(x, x')= \begin{cases} \sigma^2, \quad x = x' \\"
        r"0, \quad x \neq x'\\ \end{cases}"
    )

    _parameters_str = [("variance", r"\sigma^2")]

    def __init__(self, *, variance: float = 1):
        self.variance = variance

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        return self.variance * np.eye(x.shape[0])

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return sklearn_kern.WhiteKernel(noise_level=self.variance)


class Matern(Covariance):
    r"""
    Matérn covariance function.

    The covariance function is

    .. math::
        K(x, x') = \sigma^2 \frac{2^{1-\nu}}{\Gamma(\nu)}
        \left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)^{\nu}
        K_{\nu}\left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)

    where :math:`\sigma^2` is the variance, :math:`l` is the length scale
    and :math:`\nu` controls the smoothness of the related Gaussian process.
    The trajectories of a Gaussian process with Matérn covariance is 
    :math:`\lceil \nu \rceil - 1` times differentiable.


    Heatmap plot of the covariance function:

    .. jupyter-execute::

        from skfda.misc.covariances import Matern
        import matplotlib.pyplot as plt

        Matern().heatmap(limits=(0, 1))
        plt.show()

    Example of Gaussian process trajectories using this covariance:

    .. jupyter-execute::

        from skfda.misc.covariances import Matern
        from skfda.datasets import make_gaussian_process
        import matplotlib.pyplot as plt

        gp = make_gaussian_process(
                n_samples=10, cov=Matern(), random_state=0)
        gp.plot()
        plt.show()

    Default representation in a Jupyter notebook:

    .. jupyter-execute::

        from skfda.misc.covariances import Matern

        Matern()

    """
    _latex_formula = (
        r"K(x, x') = \sigma^2 \frac{2^{1-\nu}}{\Gamma(\nu)}"
        r"\left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)^{\nu}"
        r"K_{\nu}\left( \frac{\sqrt{2\nu}|x - x'|}{l} \right)"
    )

    _parameters_str = [
        ("variance", r"\sigma^2"),
        ("length_scale", "l"),
        ("nu", r"\nu"),
    ]

    def __init__(
        self,
        *,
        variance: float = 1,
        length_scale: float = 1,
        nu: float = 1.5,
    ) -> None:
        self.variance = variance
        self.length_scale = length_scale
        self.nu = nu

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        x = _transform_to_2d(x)
        y = _transform_to_2d(y)

        x_y_squared = _squared_norms(x, y)
        x_y = np.sqrt(x_y_squared)

        p = self.nu - 0.5
        if p.is_integer():
            # Formula for half-integers
            p = int(p)
            body = np.sqrt(2 * p + 1) * x_y / self.length_scale
            exponential = np.exp(-body)
            power_list = np.full(shape=(p,) + body.shape, fill_value=2 * body)
            power_list = np.cumprod(power_list, axis=0)
            power_list = np.concatenate(
                (power_list[::-1], np.asarray([np.ones_like(body)])),
            )
            power_list = np.moveaxis(power_list, 0, -1)
            numerator = np.cumprod(np.arange(p, 0, -1))
            numerator = np.concatenate(([1], numerator))
            denom1 = np.cumprod(np.arange(2 * p, p, -1))
            denom1 = np.concatenate((denom1[::-1], [1]))
            denom2 = np.cumprod(np.arange(1, p + 1))
            denom2 = np.concatenate(([1], denom2))

            sum_terms = power_list * numerator / (denom1 * denom2)
            return (  # type: ignore[no-any-return]
                self.variance * exponential * np.sum(sum_terms, axis=-1)
            )
        elif self.nu == np.inf:
            return (
                self.variance * np.exp(
                    -x_y_squared / (2 * self.length_scale ** 2),
                )
            )
        else:
            # General formula
            scaling = 2**(1 - self.nu) / gamma(self.nu)
            body = np.sqrt(2 * self.nu) * x_y / self.length_scale
            power = body**self.nu
            bessel = kv(self.nu, body)

            with np.errstate(invalid='ignore'):
                eval_cov = self.variance * scaling * power * bessel

            # Values with nan are where the distance is 0
            return np.nan_to_num(  # type: ignore[no-any-return]
                eval_cov,
                nan=self.variance,
            )

    def to_sklearn(self) -> sklearn_kern.Kernel:
        return (
            self.variance
            * sklearn_kern.Matern(length_scale=self.length_scale, nu=self.nu)
        )


class Empirical(Covariance):
    r"""
    Sample covariance function.

    The sample covariance function is defined as
    . math::
        K(t, s) = \frac{1}{N-\text{correction}}\sum_{n=1}^N\left(x_n(t) -
        \bar{x}(t)\right) \left(x_n(s) - \bar{x}(s)\right)

    where :math:`x_n(t)` is the n-th sample and :math:`\bar{x}(t)` is the
    mean of the samples. :math:`N` is the number of samples,
    :math:`\text{correction}` is the degrees of freedom adjustment and is such
    that :math:`N-\text{correction}` is the divisor used in the calculation of
    the covariance function.

    """

    _latex_formula = (
        r"K(t, s) = \frac{1}{N-\text{correction}}\sum_{n=1}^N"
        r"(x_n(t) - \bar{x}(t))(x_n(s) - \bar{x}(s))"
    )
    _parameters_str = [
        ("data", "data"),
    ]

    cov_fdata: FData
    correction: int

    @abc.abstractmethod
    def __init__(self, data: FData) -> None:
        if data.dim_domain != 1 or data.dim_codomain != 1:
            raise NotImplementedError(
                "Covariance only implemented "
                "for univariate functions",
            )

    def __call__(self, x: ArrayLike, y: ArrayLike) -> NDArrayFloat:
        """Evaluate the covariance function.

        Args:
            x: First array of points of evaluation.
            y: Second array of points of evaluation.

        Returns:
            Covariance function evaluated at the grid formed by x and y.
        """
        return self.cov_fdata([x, y], grid=True)[0, ..., 0]


class EmpiricalGrid(Empirical):
    """Sample covariance function for FDataGrid."""

    cov_fdata: FDataGrid
    correction: int

    def __init__(self, data: FDataGrid, correction: int = 0) -> None:
        super().__init__(data=data)

        self.correction = correction
        self.cov_fdata = data.copy(
            data_matrix=np.cov(
                data.data_matrix[..., 0],
                rowvar=False,
                ddof=correction,
            )[np.newaxis, ...],
            grid_points=[
                data.grid_points[0],
                data.grid_points[0],
            ],
            domain_range=[
                data.domain_range[0],
                data.domain_range[0],
            ],
            argument_names=data.argument_names * 2,
            sample_names=("covariance",),
        )


class EmpiricalBasis(Empirical):
    """
    Sample covariance function for FDataBasis.

    In this case one may use the basis expression of the samples to
    express the sample covariance function in the tensor product basis
    of the original basis.
    """

    cov_fdata: FDataBasis
    coeff_matrix: NDArrayFloat
    correction: int

    def __init__(self, data: FDataBasis, correction: int = 0) -> None:
        super().__init__(data=data)

        self.correction = correction
        self.coeff_matrix = np.cov(
            data.coefficients,
            rowvar=False,
            ddof=correction,
        )
        self.cov_fdata = FDataBasis(
            basis=TensorBasis([data.basis, data.basis]),
            coefficients=self.coeff_matrix.flatten(),
            argument_names=data.argument_names * 2,
            sample_names=("covariance",),
        )



from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "constants",
    ],
    submod_attrs={
        "_utils": [
            "_cartesian_product",
            "_check_array_key",
            "_check_estimator",
            "_classifier_get_classes",
            "_evaluate_grid",
            "_int_to_real",
            "_MapAcceptable",
            "_pairwise_symmetric",
            "_same_domain",
            "_to_grid",
            "_to_grid_points",
            "nquad_vec",
        ],
        '_warping': [
            "invert_warping",
            "normalize_scale",
            "normalize_warping",
        ],
    },
)

if TYPE_CHECKING:

    from ._utils import (
        _cartesian_product as _cartesian_product,
        _check_array_key as _check_array_key,
        _check_estimator as _check_estimator,
        _classifier_get_classes as _classifier_get_classes,
        _evaluate_grid as _evaluate_grid,
        _int_to_real as _int_to_real,
        _MapAcceptable as _MapAcceptable,
        _pairwise_symmetric as _pairwise_symmetric,
        _same_domain as _same_domain,
        _to_grid as _to_grid,
        _to_grid_points as _to_grid_points,
        nquad_vec as nquad_vec,
    )

    from ._warping import (
        invert_warping as invert_warping,
        normalize_scale as normalize_scale,
        normalize_warping as normalize_warping,
    )



from __future__ import annotations

from typing import Callable

import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import Kernel, WhiteKernel

from ....misc.covariances import Covariance, EmpiricalGrid
from ....representation import FDataGrid
from ....typing._numpy import NDArrayFloat
from ._empirical import EmpiricalCovariance


class ParametricGaussianCovariance(EmpiricalCovariance[FDataGrid]):

    covariance_: Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat]

    def __init__(
        self,
        cov: Kernel | Covariance,
        *,
        assume_centered: bool = False,
        fit_noise: bool = True,
    ) -> None:
        super().__init__(assume_centered=assume_centered)
        self.cov = cov
        self.fit_noise = fit_noise

    def fit(
        self,
        X: FDataGrid,
        y: object = None,
    ) -> ParametricGaussianCovariance:

        self._fit_mean(X)

        X_centered = X - self.location_

        data_matrix = X_centered.data_matrix[:, :, 0]

        grid_points = X_centered.grid_points[0][:, np.newaxis]

        cov = self.cov
        to_sklearn = getattr(cov, "to_sklearn", None)
        if to_sklearn:
            cov = to_sklearn()

        if self.fit_noise:
            cov += WhiteKernel()

        regressor = GaussianProcessRegressor(kernel=cov)
        regressor.fit(grid_points, data_matrix.T)
        self.cov_ = regressor.kernel_

        # TODO: Skip cov computation?
        # TODO: Use a user-public structure to represent the covariance,
        #  instead of a Callable object
        self.covariance_ = X.cov()
        assert isinstance(self.covariance_, EmpiricalGrid)
        self.covariance_.cov_fdata = self.covariance_.cov_fdata.copy(
            data_matrix=regressor.kernel_(
                grid_points,
            )[np.newaxis, ...],
        )

        return self



"""Typing for norms and metrics."""
from abc import abstractmethod
from typing import TypeVar

from typing_extensions import Protocol

from ._base import Vector
from ._numpy import NDArrayFloat

VectorType = TypeVar("VectorType", contravariant=True, bound=Vector)
MetricElementType = TypeVar("MetricElementType", contravariant=True)


class Norm(Protocol[VectorType]):
    """Protocol for a norm of a vector."""

    @abstractmethod
    def __call__(self, __vector: VectorType) -> NDArrayFloat:  # noqa: WPS112
        """Compute the norm of a vector."""


class Metric(Protocol[MetricElementType]):
    """Protocol for a metric between two elements of a metric space."""

    @abstractmethod
    def __call__(
        self,
        __e1: MetricElementType,  # noqa: WPS112
        __e2: MetricElementType,  # noqa: WPS112
    ) -> NDArrayFloat:
        """Compute the metric between two vectors."""



"""Evaluation Transformer Module."""
from __future__ import annotations

from typing import TypeVar, overload

from sklearn.utils.validation import check_is_fitted
from typing_extensions import Literal

from ..._utils._sklearn_adapter import BaseEstimator, InductiveTransformerMixin
from ...representation._functional_data import FData
from ...representation.extrapolation import ExtrapolationLike
from ...representation.grid import FDataGrid
from ...typing._base import GridPointsLike
from ...typing._numpy import ArrayLike, NDArrayFloat

Input = TypeVar("Input", bound=FData)
SelfType = TypeVar(
    "SelfType",
    bound="EvaluationTransformer",  # type: ignore[type-arg]
)


class EvaluationTransformer(
    BaseEstimator,
    InductiveTransformerMixin[Input, NDArrayFloat, object],
):
    r"""
    Transformer returning the evaluations of FData objects as a matrix.

    Args:
        eval_points (array_like): List of points where the functions are
            evaluated. If `None`, the functions must be `FDatagrid` objects
            and all points will be returned.
        extrapolation (str or Extrapolation, optional): Controls the
            extrapolation mode for elements outside the :term:`domain` range.
            By default it is used the mode defined during the instance of the
            object.
        grid (bool, optional): Whether to evaluate the results on a grid
            spanned by the input arrays, or at points specified by the
            input arrays. If true the eval_points should be a list of size
            dim_domain with the corresponding times for each axis. The
            return matrix has shape n_samples x len(t1) x len(t2) x ... x
            len(t_dim_domain) x dim_codomain. If the domain dimension is 1
            the parameter has no efect. Defaults to False.

    Attributes:
        shape\_ (tuple): original shape of coefficients per sample.

    Examples:
        >>> from skfda.representation import FDataGrid, FDataBasis
        >>> from skfda.preprocessing.feature_construction import (
        ...     EvaluationTransformer,
        ... )
        >>> from skfda.representation.basis import MonomialBasis

        Functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}`.

        >>> data_matrix = [[1, 2], [2, 3]]
        >>> grid_points = [2, 4]
        >>> fd = FDataGrid(data_matrix, grid_points)

        >>> transformer = EvaluationTransformer()
        >>> transformer.fit_transform(fd)
        array([[ 1., 2.],
               [ 2., 3.]])

        Functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}\longmapsto\mathbb{R}^2`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> grid_points = [2, 4]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>>
        >>> transformer = EvaluationTransformer()
        >>> transformer.fit_transform(fd)
        array([[ 1. ,  0.3,  2. ,  0.4],
               [ 2. ,  0.5,  3. ,  0.6]])

        Representation of a functional data object with 2 samples
        representing a function :math:`f : \mathbb{R}^2\longmapsto\mathbb{R}`.

        >>> data_matrix = [[[1, 0.3], [2, 0.4]], [[2, 0.5], [3, 0.6]]]
        >>> grid_points = [[2, 4], [3, 6]]
        >>> fd = FDataGrid(data_matrix, grid_points)
        >>>
        >>> transformer = EvaluationTransformer()
        >>> transformer.fit_transform(fd)
        array([[ 1. ,  0.3,  2. ,  0.4],
               [ 2. ,  0.5,  3. ,  0.6]])

        Evaluation of a functional data object at several points.

        >>> basis = MonomialBasis(n_basis=4)
        >>> coefficients = [[0.5, 1, 2, .5], [1.5, 1, 4, .5]]
        >>> fd = FDataBasis(basis, coefficients)
        >>>
        >>> transformer = EvaluationTransformer([0, 0.2, 0.5, 0.7, 1])
        >>> transformer.fit_transform(fd)
        array([[ 0.5   ,  0.784 ,  1.5625,  2.3515,  4.    ],
               [ 1.5   ,  1.864 ,  3.0625,  4.3315,  7.    ]])

    """

    @overload
    def __init__(
        self: EvaluationTransformer[FDataGrid],
        eval_points: None = None,
        *,
        extrapolation: ExtrapolationLike | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        eval_points: ArrayLike,
        *,
        extrapolation: ExtrapolationLike | None = None,
        grid: Literal[False] = False,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        eval_points: GridPointsLike,
        *,
        extrapolation: ExtrapolationLike | None = None,
        grid: Literal[True],
    ) -> None:
        pass

    def __init__(
        self,
        eval_points: ArrayLike | GridPointsLike | None = None,
        *,
        extrapolation: ExtrapolationLike | None = None,
        grid: bool = False,
    ) -> None:
        self.eval_points = eval_points
        self.extrapolation = extrapolation
        self.grid = grid

    def fit(  # noqa: D102
        self: SelfType,
        X: FData | NDArrayFloat,
        y: object = None,
    ) -> SelfType:
        if (
            callable(X)
            and self.eval_points is None
            and not isinstance(X, FDataGrid)
        ):
            raise ValueError(
                "If no eval_points are passed, the functions "
                "should be FDataGrid objects.",
            )

        self._is_fitted = True

        return self

    def transform(  # noqa: D102
        self,
        X: FData | NDArrayFloat,
        y: object = None,
    ) -> NDArrayFloat:
        check_is_fitted(self, '_is_fitted')

        if callable(X):
            if self.eval_points is None:
                assert isinstance(X, FDataGrid)
                evaluation = X.data_matrix.copy()
            else:
                evaluation = X(
                    self.eval_points,
                    extrapolation=self.extrapolation,
                    grid=self.grid,
                )
        else:
            evaluation = X

        return evaluation.reshape((len(X), -1))
